# -*- coding: utf-8 -*-

"""Base text classification model
"""

from keras.layers import *
from keras_bert import load_trained_model_from_checkpoint

from fancy_nlp.layers import NonMaskingLayer
from fancy_nlp.models.base_model import BaseModel


class BaseTextClassificationModel(BaseModel):
    def __init__(self,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2):

        self.use_char = use_char
        self.char_embeddings = char_embeddings
        self.char_vocab_size = char_vocab_size
        self.char_embed_dim = char_embed_dim
        self.char_embed_trainable = char_embed_trainable
        self.use_bert = use_bert
        self.bert_config_file = bert_config_file
        self.bert_checkpoint_file = bert_checkpoint_file
        self.bert_trainable = bert_trainable
        self.use_word = use_word
        self.word_embeddings = word_embeddings
        self.word_vocab_size = word_vocab_size
        self.word_embed_dim = word_embed_dim
        self.word_embed_trainable = word_embed_trainable
        self.max_len = max_len
        self.dropout = dropout

        assert self.use_char or self.use_bert, "must use char or bert embedding as main input"
        assert not (self.use_bert and self.max_len is None), \
            "max_len must be provided when using bert embedding as input"

    def build_input(self):
        model_inputs = []
        input_embed = []

        # TODO: consider masking
        if self.use_char:
            if self.char_embeddings is not None:
                char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                 output_dim=self.char_embed_dim,
                                                 weights=[self.char_embeddings],
                                                 trainable=self.char_embed_trainable)
            else:
                char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                 output_dim=self.char_embed_dim)
            input_char = Input(shape=(self.max_len,))
            model_inputs.append(input_char)
            input_embed.append(SpatialDropout1D(self.dropout)(char_embedding_layer(input_char)))

        if self.use_bert:
            bert_model = load_trained_model_from_checkpoint(self.bert_config_file,
                                                            self.bert_checkpoint_file,
                                                            trainable=self.bert_trainable,
                                                            output_layer_num=1,
                                                            seq_len=self.max_len)
            model_inputs.extend(bert_model.inputs)
            bert_embed = NonMaskingLayer()(bert_model.output)
            input_embed.append(SpatialDropout1D(0.2)(bert_embed))

        if self.use_word:
            if self.word_embeddings is not None:
                word_embedding_layer = Embedding(input_dim=self.word_vocab_size,
                                                 output_dim=self.word_embed_dim,
                                                 weights=[self.word_embeddings],
                                                 trainable=self.word_embed_trainable)
            else:
                word_embedding_layer = Embedding(input_dim=self.word_vocab_size,
                                                 output_dim=self.word_embed_dim)
            input_word = Input(shape=(self.max_len,))
            model_inputs.append(input_word)
            input_embed.append(SpatialDropout1D(self.dropout)(word_embedding_layer(input_word)))

        input_embed = concatenate(input_embed) if len(input_embed) > 1 else input_embed[0]
        return model_inputs, input_embed

    def build_model(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-

import tensorflow as tf
from keras.layers import *
from keras.models import Model
from keras.optimizers import Adam

from fancy_nlp.models.spm.base_spm_model import BaseSPMModel
from fancy_nlp.layers.matching import *


class SiameseCNN(BaseSPMModel):
    """Siamese CNN model for SPM.
    """
    def __init__(self,
                 num_class,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 max_len=None,
                 max_word_len=None,
                 dropout=0.2,
                 filters=200,
                 kernel_size=(2, 3, 4, 5),
                 char_dim=50,
                 fc_dim=200,
                 activation='relu',
                 optimizer='adam'):
        self.num_class = num_class
        self.filters = filters
        self.kernel_size = kernel_size
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(SiameseCNN, self).__init__(use_word, word_embeddings, word_vocab_size,
                                         word_embed_dim, word_embed_trainable, use_char,
                                         char_embeddings, char_vocab_size, char_embed_dim,
                                         char_embed_trainable, use_bert, bert_config_file,
                                         bert_checkpoint_file, bert_trainable,
                                         bert_output_layer_num, False, max_len,
                                         max_word_len, char_dim, dropout)

    def build_model(self):
        model_inputs, input_embed_a, input_embed_b = self.build_input()

        a_conv_layers = []
        b_conv_layers = []
        for filter_length in self.kernel_size:
            conv_layer = Conv1D(filters=self.filters, kernel_size=filter_length, padding='valid',
                                activation='relu', strides=1)
            a_conv = conv_layer(input_embed_a)
            b_conv = conv_layer(input_embed_b)
            a_conv_layers.append(GlobalMaxPooling1D()(a_conv))
            b_conv_layers.append(GlobalMaxPooling1D()(b_conv))
        a_conv = concatenate(inputs=a_conv_layers)
        b_conv = concatenate(inputs=b_conv_layers)
        sent_rep = concatenate([a_conv, b_conv])
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model


class SiameseBiLSTM(BaseSPMModel):
    """Siamese Bidirectional LSTM model for SPM.
    Support using CUDANNLSTM for acceleration when gpu is available.
    """
    def __init__(self,
                 num_class,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 max_len=None,
                 max_word_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 char_dim=50,
                 fc_dim=200,
                 activation='relu',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(SiameseBiLSTM, self).__init__(use_word, word_embeddings, word_vocab_size,
                                            word_embed_dim, word_embed_trainable, use_char,
                                            char_embeddings, char_vocab_size, char_embed_dim,
                                            char_embed_trainable, use_bert, bert_config_file,
                                            bert_checkpoint_file, bert_trainable,
                                            bert_output_layer_num, False, max_len,
                                            max_word_len, char_dim, dropout)

    def build_model(self):
        model_inputs, input_embed_a, input_embed_b = self.build_input()

        if tf.test.is_gpu_available(cuda_only=True):
            bilstm_layer = Bidirectional(CuDNNLSTM(self.rnn_units))
        else:
            bilstm_layer = Bidirectional(LSTM(self.rnn_units))
        a_lstm = bilstm_layer(input_embed_a)
        b_lstm = bilstm_layer(input_embed_b)

        sent_rep = concatenate([a_lstm, b_lstm])
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model


class SiameseBiGRU(BaseSPMModel):
    """Siamese Bidirectional GRU model for SPM.
    Support using CUDANNGRU for acceleration when gpu is available.
    """

    def __init__(self,
                 num_class,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 max_len=None,
                 max_word_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 char_dim=50,
                 fc_dim=200,
                 activation='relu',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(SiameseBiGRU, self).__init__(use_word, word_embeddings, word_vocab_size,
                                           word_embed_dim, word_embed_trainable, use_char,
                                           char_embeddings, char_vocab_size, char_embed_dim,
                                           char_embed_trainable, use_bert, bert_config_file,
                                           bert_checkpoint_file, bert_trainable,
                                           bert_output_layer_num, False, max_len,
                                           max_word_len, char_dim, dropout)

    def build_model(self):
        model_inputs, input_embed_a, input_embed_b = self.build_input()

        if tf.test.is_gpu_available(cuda_only=True):
            bigru_layer = Bidirectional(CuDNNGRU(self.rnn_units))
        else:
            bigru_layer = Bidirectional(GRU(self.rnn_units))
        a_gru = bigru_layer(input_embed_a)
        b_gru = bigru_layer(input_embed_b)

        sent_rep = concatenate([a_gru, b_gru])
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model


class ESIM(BaseSPMModel):
    """Enhanced LSTM model (esim) for SPM.
    Support using CUDANNLSTM for acceleration when gpu is available.
    """
    def __init__(self,
                 num_class,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 max_len=None,
                 max_word_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 char_dim=50,
                 fc_dim=200,
                 activation='relu',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(ESIM, self).__init__(use_word, word_embeddings, word_vocab_size,
                                   word_embed_dim, word_embed_trainable, use_char,
                                   char_embeddings, char_vocab_size,  char_embed_dim,
                                   char_embed_trainable, use_bert, bert_config_file,
                                   bert_checkpoint_file, bert_trainable,
                                   bert_output_layer_num, False, max_len,
                                   max_word_len, char_dim, dropout)

    def build_model(self):
        model_inputs, input_embed_a, input_embed_b = self.build_input()

        if tf.test.is_gpu_available(cuda_only=True):
            bilstm_layer = Bidirectional(CuDNNLSTM(self.rnn_units, return_sequences=True))
        else:
            bilstm_layer = Bidirectional(LSTM(self.rnn_units, return_sequences=True))
        a_lstm = bilstm_layer(input_embed_a)
        b_lstm = bilstm_layer(input_embed_b)

        # interaction attention
        attention = Dot(axes=-1)([a_lstm, b_lstm])

        wb = Lambda(lambda x: K.softmax(x, axis=1), output_shape=lambda x: x)(attention)
        wa = Permute((2, 1))(Lambda(lambda x: K.softmax(x, axis=2), output_shape=lambda x: x)(attention))
        a_ = Dot(axes=1)([wa, b_lstm])
        b_ = Dot(axes=1)([wb, a_lstm])

        substract_a = Subtract()([a_lstm, a_])
        multiply_a = Multiply()([a_lstm, a_])
        substract_b = Subtract()([b_lstm, b_])
        multiply_b = Multiply()([b_lstm, b_])

        m_a = concatenate([a_lstm, a_, substract_a, multiply_a], axis=-1)
        m_b = concatenate([b_lstm, b_, substract_b, multiply_b], axis=-1)

        if tf.test.is_gpu_available(cuda_only=True):
            compose = Bidirectional(CuDNNLSTM(self.rnn_units, return_sequences=True))
        else:
            compose = Bidirectional(LSTM(self.rnn_units, return_sequences=True))
        v_a = compose(m_a)
        v_b = compose(m_b)

        a_maxpool = GlobalMaxPool1D()(v_a)
        b_maxpool = GlobalMaxPool1D()(v_b)
        a_avgpool = GlobalAvgPool1D()(v_a)
        b_avgpool = GlobalAvgPool1D()(v_b)
        a_pool = concatenate([a_avgpool, a_maxpool], axis=-1)
        b_pool = concatenate([b_avgpool, b_maxpool], axis=-1)

        sent_rep = concatenate([a_pool, b_pool])
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model


class BiMPM(BaseSPMModel):
    """Bilateral Multi-perspective Matching (BiMPM) model for SPM.
    Support using CUDANNLSTM for acceleration when gpu is available.
    """

    def __init__(self,
                 num_class,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 max_len=None,
                 max_word_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 char_dim=50,
                 fc_dim=200,
                 activation='relu',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(BiMPM, self).__init__(use_word, word_embeddings, word_vocab_size,
                                    word_embed_dim, word_embed_trainable, use_char,
                                    char_embeddings, char_vocab_size, char_embed_dim,
                                    char_embed_trainable, use_bert, bert_config_file,
                                    bert_checkpoint_file, bert_trainable,
                                    bert_output_layer_num, False, max_len, max_word_len,
                                    char_dim, dropout)

    def build_model(self):
        model_inputs, input_embed_a, input_embed_b = self.build_input()

        if tf.test.is_gpu_available(cuda_only=True):
            bilstm_layer = Bidirectional(CuDNNLSTM(self.rnn_units, return_sequences=True))
        else:
            bilstm_layer = Bidirectional(LSTM(self.rnn_units, return_sequences=True))
        a_lstm = bilstm_layer(input_embed_a)
        b_lstm = bilstm_layer(input_embed_b)

        matching1 = FullMatching()
        a_matching1 = matching1([a_lstm, GlobalAvgPool1D()(b_lstm)])
        b_matching1 = matching1([b_lstm, GlobalAvgPool1D()(a_lstm)])

        matching2_forward = MaxPoolingMatching()
        a_matching2 = matching2_forward([a_lstm, b_lstm])
        b_matching2 = matching2_forward([b_lstm, a_lstm])

        cos = Dot(axes=-1, normalize=True)([a_lstm, b_lstm])
        wb = Lambda(lambda x: K.softmax(x, axis=1), output_shape=lambda x: x)(cos)
        wa = Permute((2, 1))(Lambda(lambda x: K.softmax(x, axis=2), output_shape=lambda x: x)(cos))

        a_ = Dot(axes=1)([wa, b_lstm])
        b_ = Dot(axes=1)([wb, a_lstm])

        matching3 = AttentiveMatching(perspective_num=5)
        a_matching3 = matching3([a_lstm, a_])
        b_matching3 = matching3([b_lstm, b_])

        a_max = Lambda(
            lambda x: K.batch_dot(K.one_hot(K.argmax(x[0], axis=1), K.int_shape(x[1])[-2]), x[1]),
            output_shape=lambda x: x[1])([wb, a_lstm])
        b_max = Lambda(
            lambda x: K.batch_dot(K.one_hot(K.argmax(x[0], axis=1), K.int_shape(x[1])[-2]), x[1]),
            output_shape=lambda x: x[1])([wa, b_lstm])

        matching4 = MaxAttentiveMatching(perspective_num=5)
        a_matching4 = matching4([a_lstm, b_max])
        b_matching4 = matching4([b_lstm, a_max])

        a_matching = concatenate([a_matching1, a_matching2, a_matching3, a_matching4])
        b_matching = concatenate([b_matching1, b_matching2, b_matching3, b_matching4])

        if tf.test.is_gpu_available(cuda_only=True):
            aggregation_layer = Bidirectional(CuDNNLSTM(self.rnn_units))
        else:
            aggregation_layer = Bidirectional(LSTM(self.rnn_units))
        a_bimpm = aggregation_layer(a_matching)
        b_bimpm = aggregation_layer(b_matching)

        sent_rep = concatenate([a_bimpm, b_bimpm])
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model


class BertSPM(BaseSPMModel):
    """Bert model for SPM.
    We suggest you to train bert on machines with GPU cause it will be very slow to be trained with
    cpu. You will have to re-install a gpu version of tensorflow to do so.
    """

    def __init__(self,
                 num_class,
                 bert_config_file,
                 bert_checkpoint_file,
                 bert_trainable,
                 max_len,
                 dropout=0.2,
                 fc_dim=100,
                 activation='tanh',
                 optimizer=Adam(lr=1e-5),  # use a small learning rate for bert
                 bert_output_layer_num=1,
                 **kwargs):
        self.num_class = num_class
        self.fc_dim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(BertSPM, self).__init__(use_word=False, use_char=False, use_bert=True,
                                      bert_config_file=bert_config_file,
                                      bert_checkpoint_file=bert_checkpoint_file,
                                      bert_trainable=bert_trainable,
                                      bert_output_layer_num=bert_output_layer_num,
                                      use_bert_model=True,
                                      max_len=max_len, dropout=dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()

        sent_rep = Lambda(lambda x: x[:, 0])(input_embed)
        sent_rep = Dense(self.fc_dim, activation=self.activation)(sent_rep)
        matching_score = Dense(self.num_class, activation='softmax')(sent_rep)

        spm_loss = 'categorical_crossentropy'
        spm_metrics = 'accuracy'
        spm_model = Model(model_inputs, matching_score)
        spm_model.compile(optimizer=self.optimizer, loss=spm_loss, metrics=[spm_metrics])
        return spm_model



# -*- coding: utf-8 -*-

from absl import logging
from keras.preprocessing.sequence import pad_sequences

from ..utils import load_pre_trained, train_w2v, train_fasttext


class Preprocessor(object):
    """Base Preprocessor.
    """

    def __init__(self,
                 max_len=None,
                 padding_mode='post',
                 truncating_mode='post'):
        self.pad_token = '<PAD>'
        self.unk_token = '<UNK>'
        self.cls_token = '<CLS>'
        self.seq_token = '<SEQ>'

        self.max_len = max_len
        self.padding_mode = padding_mode
        self.truncating_mode = truncating_mode

    @staticmethod
    def build_corpus(untokenized_texts, cut_func):
        """Build corpus from untokenized texts.

        Args:
            cut_func: function to tokenize texts. For example, lambda x: list(x) can be used for
                      tokenize texts in char level, while lambda x: jieba.lcut(x) can be used for
                      tokenize Chinese texts in word level.

        Returns: list of tokenized texts, like ``[['我', '是', '中', '国', '人']]``

        """
        corpus = []
        for text in untokenized_texts:
            corpus.append(cut_func(text))
        return corpus

    def build_vocab(self, corpus, min_count=3, special_token='standard'):
        """Build vocabulary using corpus.

        Args:
            corpus: list of tokenized texts, like ``[['我', '是', '中', '国', '人']]``
            min_count: token of which frequency is less than min_count will be ignored
            special_token: str, how to handle special tokens. If special_token is 'standard', we
                           add 2 special tokens: [('<PAD>', 0), ('<UNK>', 1)]. If special_token is
                           'bert', we add 4 special tokens: [('<PAD>', 0), ('<UNK>', 1),
                           ('<CLS>', 2), ('<SEQ>', 3)]

        Returns: tuple(dict, dict, dict):
                 1. token_count: a mapping of tokens to frequencies
                 2. token_vocab: a mapping of tokens to indices
                 3. id2token: a mapping of indices to tokens

        """
        if special_token == 'standard':
            token_vocab = {self.pad_token: 0,
                           self.unk_token: 1}
        elif special_token == 'bert':
            token_vocab = {self.pad_token: 0,
                           self.unk_token: 1,
                           self.cls_token: 2,
                           self.seq_token: 3}
        else:
            raise ValueError('Argument `special_token` can only be "standard" or "bert", '
                             'got: {}'.format(special_token))

        token_count = {}
        for tokenized_text in corpus:
            for token in tokenized_text:
                token_count[token] = token_count.get(token, 0) + 1
        # filter out low-frequency token
        token_count = {token: count for token, count in token_count.items()
                       if count >= min_count}

        for token in token_count:
            token_vocab[token] = len(token_vocab)
        id2token = dict((idx, token) for token, idx in token_vocab.items())

        logging.info('Build vocabulary finished, vocabulary size: {}'.format(len(token_vocab)))
        return token_count, token_vocab, id2token

    def build_label_vocab(self, labels):
        """Build label vocabulary
        """
        raise NotImplementedError

    def build_embedding(self, embed_type, vocab, corpus=None, embedding_dim=300,
                        special_token='standard'):
        """Prepare embeddings for the words in vocab.
        We support loading external pre-trained embeddings as well as training on the corpus to
        obtain embeddings

        Args:
            embed_type: str, can be a path to pre-trained embedding file or pre-train embedding
                        method to train on corpus
            vocab: a mapping of words to indices
            corpus: a list of tokenized texts
            embedding_dim: dimensionality of embedding
            special_token: str, how to handle special tokens. If special_token is 'standard', we
                           add 2 special tokens: [('<PAD>', 0), ('<UNK>', 1)]. If special_token is
                           'bert', we add 4 special tokens: [('<PAD>', 0), ('<UNK>', 1),
                           ('<CLS>', 2), ('<SEQ>', 3)].
                           We will use zero-initializer for '<PAD>' token and random-initializer
                           for other special tokens.
        """
        zero_init_indices = vocab.get(self.pad_token)
        if special_token == 'standard':
            rand_init_indices = vocab.get(self.unk_token)
        elif special_token == 'bert':
            rand_init_indices = [vocab.get(self.unk_token),
                                 vocab.get(self.cls_token),
                                 vocab.get(self.seq_token)]
        else:
            raise ValueError('Argument `special_token` can only be "standard" or "bert", '
                             'got: {}'.format(special_token))

        if embed_type is None:
            return None     # do not adopt any pre-trained embeddings
        if embed_type == 'word2vec':
            return train_w2v(corpus, vocab, zero_init_indices, rand_init_indices, embedding_dim)
        elif embed_type == 'fasttext':
            return train_fasttext(corpus, vocab, zero_init_indices, rand_init_indices,
                                  embedding_dim)
        else:
            try:
                return load_pre_trained(embed_type, embedding_dim, vocab,
                                        zero_init_indices, rand_init_indices)
            except FileNotFoundError:
                raise ValueError('Argument `embed_type` input error: {}'.format(embed_type))

    def prepare_input(self, data, label=None):
        """Prepare input for neural model training, evaluating and testing
        """
        raise NotImplementedError

    @staticmethod
    def build_id_sequence(tokenized_text, vocabulary, unk_idx=1):
        """Given a token list, return the corresponding id sequence.

        Args:
            tokenized_text: list of str, like `['我', '是', '中', '国', '人']`
            vocabulary: a mapping of words to indices
            unk_idx: the index of words that do not appear in vocabulary, we usually set it to 1

        Returns: list of indices

        """
        return [vocabulary.get(token, unk_idx) for token in tokenized_text]

    @staticmethod
    def build_id_matrix(tokenized_texts, vocabulary, unk_idx=1):
        """Given a list, each item is a token list, return the corresponding id matrix.

        Args:
            tokenized_texts: list of tokenized texts, like ``[['我', '是', '中', '国', '人']]``
            vocabulary: a mapping of words to indices
            unk_idx: the index of words that do not appear in vocabulary, we usually set it to 1

        Returns: list of list of indices

        """
        return [[vocabulary.get(token, unk_idx) for token in text] for text in tokenized_texts]

    def pad_sequence(self, sequence_list):
        """Given a list, each item is a id sequence, return the padded sequence
        """
        return pad_sequences(sequence_list, maxlen=self.max_len, padding=self.padding_mode,
                             truncating=self.truncating_mode)

    def label_decode(self, predictions, label_dict):
        """Decode model predictions to labels
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

import tensorflow as tf
from keras.layers import *
from keras.models import Model

from fancy_nlp.models.text_classification.base_text_classification_model import \
    BaseTextClassificationModel


class CNNTextClassification(BaseTextClassificationModel):
    """CNN model for text classification.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 fc_dim=100,
                 activation='tanh',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_fim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(CNNTextClassification, self).__init__(
            use_char, char_embeddings, char_vocab_size, char_embed_dim,
            char_embed_trainable, use_bert, bert_config_file,
            bert_checkpoint_file, bert_trainable, use_word,
            word_embeddings, word_vocab_size, word_embed_dim,
            word_embed_trainable, max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        filter_lengths = [3, 4, 5]
        conv_layers = []
        for filter_length in filter_lengths:
            conv_layer = Conv1D(filters=128, kernel_size=filter_length, padding='valid',
                                activation='relu', strides=1)(input_embed)
            conv_layers.append(conv_layer)
        poolings = [GlobalMaxPooling1D()(conv) for conv in conv_layers]
        x = Concatenate()(poolings)
        output_layer = Dense(self.num_class, activation='softmax')(x)
        text_classification_loss = 'categorical_crossentropy'
        text_classification_metrics = 'accuracy'
        text_classification_model = Model(
            model_inputs if len(model_inputs) > 1 else model_inputs[0], output_layer)
        text_classification_model.compile(
            optimizer=self.optimizer, loss=text_classification_loss,
            metrics=[text_classification_metrics])
        return text_classification_model


class RCNNTextClassification(BaseTextClassificationModel):
    """RCNN model for text classification.
       1. Support using CuDNNLSTM for acceleration when gpu is available. You will have to install
          a gpu version of tensorflow to do so.
       2. Note that model using CuDNNLSTM can only be deployed to machines with GPU for practical
          use.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 fc_dim=100,
                 activation='tanh',
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_fim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(RCNNTextClassification, self).__init__(
            use_char, char_embeddings, char_vocab_size, char_embed_dim,
            char_embed_trainable, use_bert, bert_config_file,
            bert_checkpoint_file, bert_trainable, use_word,
            word_embeddings, word_vocab_size, word_embed_dim,
            word_embed_trainable, max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        if tf.test.is_gpu_available(cuda_only=True):
            input_encode = Bidirectional(
                CuDNNLSTM(self.rnn_units, return_sequences=True))(input_embed)
        else:
            input_encode = Bidirectional(
                LSTM(self.rnn_units, return_sequences=True))(input_embed)
        x = Concatenate()([input_embed, input_encode])
        convs = []
        for kernel_size in range(1, 5):
            conv = Conv1D(128, kernel_size, activation='relu')(x)
            convs.append(conv)
        poolings = [GlobalAveragePooling1D()(conv) for conv in convs] + [GlobalMaxPooling1D()(conv)
                                                                         for conv in convs]
        x = Concatenate()(poolings)
        output_layer = Dense(self.num_class, activation='softmax')(x)

        text_classification_loss = 'categorical_crossentropy'
        text_classification_metrics = 'accuracy'
        text_classification_model = Model(
            model_inputs if len(model_inputs) > 1 else model_inputs[0], output_layer)
        text_classification_model.compile(
            optimizer=self.optimizer, loss=text_classification_loss,
            metrics=[text_classification_metrics])
        return text_classification_model


class BertTextClassification(BaseTextClassificationModel):
    """Bert model for text classification.
    We suggest you to train bert on machines with GPU cause it will be very slow to be trained with
    cpu. You will have to re-install a gpu version of tensorflow to do so.
    """

    def __init__(self,
                 num_class,
                 bert_config_file,
                 bert_checkpoint_file,
                 bert_trainable,
                 max_len,
                 dropout=0.2,
                 fc_dim=100,
                 activation='tanh',
                 optimizer='adam'):
        self.num_class = num_class
        self.fc_fim = fc_dim
        self.activation = activation
        self.optimizer = optimizer
        super(BertTextClassification, self).__init__(
            use_char=False, use_bert=True,
            bert_config_file=bert_config_file,
            bert_checkpoint_file=bert_checkpoint_file,
            bert_trainable=bert_trainable, use_word=False,
            max_len=max_len, dropout=dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        x = GlobalAveragePooling1D()(input_embed)

        output_layer = Dense(self.num_class, activation='softmax')(x)
        text_classification_loss = 'categorical_crossentropy'
        text_classification_metrics = 'accuracy'
        text_classification_model = Model(
            model_inputs if len(model_inputs) > 1 else model_inputs[0], output_layer)
        text_classification_model.compile(
            optimizer=self.optimizer, loss=text_classification_loss,
            metrics=[text_classification_metrics])
        return text_classification_model



# -*- coding: utf-8 -*-

import codecs

from sklearn.model_selection import train_test_split


def load_ner_data_and_labels(filename, delimiter='\t', split=False, split_size=0.1, seed=42):
    """Load ner data and label from a file.

    The file should follow CoNLL format:
    Each line is a token and its label separated by 'delimiter', or a blank line indicating the end of
    a sentence.

    Args:
        filename: str, path to ner file
        delimiter: str, delimiter to split token and label
        split: bool, whether to split into train and test subsets
        split_size: float, the proportion of test subset, between 0.0 and 1.0
        seed: int, random seed

    Returns: If split: tuple(list, list, list, list), train data and labels as well as test data and
             labels.
             Otherwise: tuple(list, list), data and labels

    """
    with codecs.open(filename, 'r', encoding='utf8') as reader:
        token_seqs, label_seqs = [], []
        tokens, labels = [], []
        for i, line in enumerate(reader):
            line = line.rstrip()
            if line:
                line_split = line.split(delimiter)
                if len(line_split) == 2:
                    token, label = line_split
                    tokens.append(token)
                    labels.append(label)
                else:
                    raise Exception(f'Format Error at line {i}!'
                                    f' Input file should follow CoNLL format.')
            else:
                if tokens:
                    token_seqs.append(tokens)
                    label_seqs.append(labels)
                    tokens, labels = [], []

        if tokens:  # in case there's no blank line at the end of the file
            token_seqs.append(tokens)
            label_seqs.append(labels)

    if split:
        x_train, x_test, y_train, y_test = train_test_split(token_seqs, label_seqs,
                                                            test_size=split_size, random_state=seed)
        return x_train, y_train, x_test, y_test
    else:
        return token_seqs, label_seqs


def load_text_classification_data_and_labels(
        filename, label_index=0, text_index=1, delimiter='\t', split_mode=0, split_size=0.1,
        use_header=False, seed=42):
    """Load text classification data and label from a file.

    Args:
        filename: str, path to ner file
        delimiter: str, delimiter to split token and label
        label_index: int, refer to which column store the classification label
        text_index:int, refer to which column store the text information
        split_mode: int, if `split_mode` is 1, it will split the dataset into train and valid;
                         if `split_mode` is 2, it will split the dataset into train, valid and test

        split_size: float, the proportion of test subset, between 0.0 and 1.0
        use_header: bool, whether to use first line as the header
        seed: int, random seed

    Returns: If split: tuple(list, list, list, list), train data and labels as well as test data and
             labels.
             Otherwise: tuple(list, list), data and labels

    """
    with codecs.open(filename, 'r', encoding='utf8') as reader:
        token_seqs, label_seqs = [], []
        for i, line in enumerate(reader):
            if use_header and i == 0:
                continue
            line_items = line.strip().split(delimiter)

            text, label = line_items[text_index], line_items[label_index]
            token_seqs.append(list(text))
            label_seqs.append(label)

    if split_mode == 1:
        x_train, x_valid, y_train, y_valid = train_test_split(
            token_seqs, label_seqs, test_size=split_size, stratify=label_seqs, random_state=seed)
        return x_train, y_train, x_valid, y_valid
    elif split_mode == 2:
        x_train, x_holdout, y_train, y_holdout = train_test_split(
            token_seqs, label_seqs, test_size=split_size, stratify=label_seqs, random_state=seed)
        x_valid, x_test, y_valid, y_test = train_test_split(
            x_holdout, y_holdout, test_size=0.5, stratify=y_holdout, random_state=seed)
        return x_train, y_train, x_valid, y_valid, x_test, y_test
    else:
        return token_seqs, label_seqs


def load_spm_data_and_labels(filename, delimiter='\t', split_mode=0, split_size=0.2, seed=42):
    """Load spm data and label from a file.

    The file should follow fixed format:
    Each line is a pair of sentences and its label separated by tab, like text_a \t text_b \t label.

    Args:
        filename: str, path to spm file
        delimiter: str, delimiter to split texts and label
        split_mode: int, if `split_mode` is 1, it will split the dataset into train and valid;
                         if `split_mode` is 2, it will split the dataset into train, valid and test
        split_size: float, the proportion of test subset, between 0.0 and 1.0
        seed: int, random seed

    Returns: If split: tuple((list_a, list_b), list, (list_a, list_b), list), train data pairs and
             labels as well as test data pairs and labels.
             Otherwise: tuple((list_a, list_b), list), data pairs and labels

    """
    with codecs.open(filename, 'r', encoding='utf8') as reader:
        text_a, text_b, labels = [], [], []
        for i, line in enumerate(reader):
            line = line.rstrip()
            line_split = line.split(delimiter)
            if line:
                if len(line_split) == 3:
                    a, b, label = line_split
                    text_a.append(a)
                    text_b.append(b)
                    labels.append(label)
                else:
                    raise Exception(f'Format Error at line {i}!'
                                    f'Input file should follow fixed spm format.')
    if split_mode == 1:
        x_a_train, x_a_valid, x_b_train, x_b_valid, y_train, y_valid = train_test_split(
            text_a, text_b, labels, test_size=split_size, stratify=labels, random_state=seed)
        return (x_a_train, x_b_train), y_train, (x_a_valid, x_b_valid), y_valid
    elif split_mode == 2:
        x_a_train, x_a_holdout, x_b_train, x_b_holdout, y_train, y_holdout = train_test_split(
            text_a, text_b, labels, test_size=split_size, stratify=labels, random_state=seed)
        x_a_valid, x_a_test, x_b_valid, x_b_test, y_valid, y_test = train_test_split(
            x_a_holdout, x_b_holdout, y_holdout, test_size=0.5, stratify=y_holdout,
            random_state=seed)
        return (x_a_train, x_b_train), y_train, (x_a_valid, x_b_valid), y_valid, \
               (x_a_test, x_b_test), y_test
    else:
        return (text_a, text_b), labels



# -*- coding: utf-8 -*-

"""Base SPM model
"""

from keras.layers import *
from keras_bert import load_trained_model_from_checkpoint

from fancy_nlp.layers import NonMaskingLayer
from fancy_nlp.models.base_model import BaseModel


class BaseSPMModel(BaseModel):
    def __init__(self,
                 use_word=True,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 use_char=False,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 bert_output_layer_num=1,
                 use_bert_model=False,
                 max_len=None,
                 max_word_len=None,
                 char_dim=50,
                 dropout=0.2):

        self.use_word = use_word
        self.word_embeddings = word_embeddings
        self.word_vocab_size = word_vocab_size
        self.word_embed_dim = word_embed_dim
        self.word_embed_trainable = word_embed_trainable
        self.use_char = use_char
        self.char_embeddings = char_embeddings
        self.char_vocab_size = char_vocab_size
        self.char_embed_dim = char_embed_dim
        self.char_embed_trainable = char_embed_trainable
        self.use_bert = use_bert
        self.bert_config_file = bert_config_file
        self.bert_checkpoint_file = bert_checkpoint_file
        self.bert_trainable = bert_trainable
        self.bert_output_layer_num = bert_output_layer_num
        self.use_bert_model = use_bert_model
        self.max_len = max_len
        self.max_word_len = max_word_len
        self.char_dim = char_dim
        self.dropout = dropout

        assert not (self.use_bert_model and (self.use_word or self.use_char)), \
            "bert model can not add word or char embedding as additional input"
        assert not (self.use_bert_model and not use_bert), "bert model must use bert" \
                                                           "embedding"
        assert self.use_word or self.use_char or self.use_bert, \
            "must use word or char or bert embedding as input"
        assert not (self.use_word and self.use_bert), \
            "bert embedding can not be used with word embedding"
        assert not (self.use_bert and self.max_len is None), \
            "max_len must be provided when using bert embedding as input"

    def build_char_embedding(self, char_embedding_layer, input_char_a, input_char_b):
        char_embedding_a = TimeDistributed(char_embedding_layer)(input_char_a)
        char_embedding_b = TimeDistributed(char_embedding_layer)(input_char_b)
        conv_layer = Conv1D(filters=self.char_dim, kernel_size=2, padding='same',
                            activation='relu', strides=1)
        char_embedding_a = TimeDistributed(conv_layer)(char_embedding_a)
        char_embedding_a = TimeDistributed(GlobalMaxPooling1D())(char_embedding_a)
        char_embedding_b = TimeDistributed(conv_layer)(char_embedding_b)
        char_embedding_b = TimeDistributed(GlobalMaxPooling1D())(char_embedding_b)
        return char_embedding_a, char_embedding_b

    def build_input(self):

        # TODO: consider masking
        if self.use_bert_model:
            model_inputs = []
            bert_model = load_trained_model_from_checkpoint(self.bert_config_file,
                                                            self.bert_checkpoint_file,
                                                            trainable=self.bert_trainable,
                                                            output_layer_num=self.bert_output_layer_num,
                                                            seq_len=self.max_len)
            input_bert = Input(shape=(self.max_len,))
            input_seg = Input(shape=(self.max_len,))
            model_inputs.append(input_bert)
            model_inputs.append(input_seg)
            bert_embed = NonMaskingLayer()(bert_model([input_bert, input_seg]))
            input_embed = SpatialDropout1D(self.dropout)(bert_embed)

            return model_inputs, input_embed

        model_inputs_a = []
        input_embed_a = []
        model_inputs_b = []
        input_embed_b = []

        if self.use_word:
            if self.word_embeddings is not None:
                word_embedding_layer = Embedding(input_dim=self.word_vocab_size,
                                                 output_dim=self.word_embed_dim,
                                                 weights=[self.word_embeddings],
                                                 trainable=self.word_embed_trainable)
            else:
                word_embedding_layer = Embedding(input_dim=self.word_vocab_size,
                                                 output_dim=self.word_embed_dim)
            input_word_a = Input(shape=(self.max_len,))
            model_inputs_a.append(input_word_a)
            input_embed_a.append(SpatialDropout1D(self.dropout)(word_embedding_layer(input_word_a)))
            input_word_b = Input(shape=(self.max_len,))
            model_inputs_b.append(input_word_b)
            input_embed_b.append(SpatialDropout1D(self.dropout)(word_embedding_layer(input_word_b)))

            if self.use_char:
                if self.char_embeddings is not None:
                    char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                     output_dim=self.char_embed_dim,
                                                     weights=[self.char_embeddings],
                                                     trainable=self.char_embed_trainable)
                else:
                    char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                     output_dim=self.char_embed_dim)
                input_char_a = Input(shape=(self.max_len, self.max_word_len))
                model_inputs_a.append(input_char_a)
                input_char_b = Input(shape=(self.max_len, self.max_word_len))
                model_inputs_b.append(input_char_b)
                char_embed_a, char_embed_b = self.build_char_embedding(char_embedding_layer,
                                                                       input_char_a,
                                                                       input_char_b)
                input_embed_a.append(SpatialDropout1D(self.dropout)(char_embed_a))
                input_embed_b.append(SpatialDropout1D(self.dropout)(char_embed_b))

        else:
            if self.use_char:
                if self.char_embeddings is not None:
                    char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                     output_dim=self.char_embed_dim,
                                                     weights=[self.char_embeddings],
                                                     trainable=self.char_embed_trainable)
                else:
                    char_embedding_layer = Embedding(input_dim=self.char_vocab_size,
                                                     output_dim=self.char_embed_dim)
                input_char_a = Input(shape=(self.max_len,))
                model_inputs_a.append(input_char_a)
                input_embed_a.append(SpatialDropout1D(self.dropout)
                                     (char_embedding_layer(input_char_a)))
                input_char_b = Input(shape=(self.max_len,))
                model_inputs_b.append(input_char_b)
                input_embed_b.append(SpatialDropout1D(self.dropout)
                                     (char_embedding_layer(input_char_b)))

            if self.use_bert:
                bert_model = load_trained_model_from_checkpoint(self.bert_config_file,
                                                                self.bert_checkpoint_file,
                                                                trainable=self.bert_trainable,
                                                                output_layer_num=self.bert_output_layer_num,
                                                                seq_len=self.max_len)
                input_bert_a = Input(shape=(self.max_len,))
                input_seg_a = Input(shape=(self.max_len,))
                model_inputs_a.append(input_bert_a)
                model_inputs_a.append(input_seg_a)
                bert_embed_a = NonMaskingLayer()(bert_model([input_bert_a, input_seg_a]))
                input_embed_a.append(SpatialDropout1D(self.dropout)(bert_embed_a))

                input_bert_b = Input(shape=(self.max_len,))
                input_seg_b = Input(shape=(self.max_len,))
                model_inputs_b.append(input_bert_b)
                model_inputs_b.append(input_seg_b)
                bert_embed_b = NonMaskingLayer()(bert_model([input_bert_b, input_seg_b]))
                input_embed_b.append(SpatialDropout1D(self.dropout)(bert_embed_b))

        input_embed_a = concatenate(input_embed_a) if len(input_embed_a) > 1 \
            else input_embed_a[0]
        input_embed_b = concatenate(input_embed_b) if len(input_embed_b) > 1 \
            else input_embed_b[0]
        return model_inputs_a + model_inputs_b, input_embed_a, input_embed_b

    def build_model(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-

import os

CACHE_DIR = '~/.fancy_cache'
if not os.path.exists(os.path.expanduser(CACHE_DIR)):
    os.makedirs(os.path.expanduser(CACHE_DIR))

MODEL_STORAGE_PREFIX = \
    'https://fancy-nlp-1253403094.cos.ap-shanghai.myqcloud.com/pretrained_models/'



# -*- coding: utf-8 -*-

from absl import logging
import numpy as np
from gensim.models import Word2Vec
from gensim.models import KeyedVectors
from gensim.models import FastText


def load_glove_format(filename, embedding_dim):
    """Load pre-trained embedding from a file in glove-embedding-like format

    Args:
        filename: str, file path to pre-trained embedding
        embedding_dim: int, dimensionality of embedding
    Returns:
        word_vector: dict(str, np.array), a mapping of words to embeddings;

    """
    word_vectors = {}
    with open(filename, 'r') as reader:
        for i, line in enumerate(reader):
            line = line.strip().split()
            word = line[0]
            word_vector = np.array([float(v) for v in line[1:]])

            if word_vector.shape[0] != embedding_dim:
                raise ValueError(f'Format error at line {i}! The size of word embedding dose not '
                                 f'equal {embedding_dim}')

            word_vectors[word] = word_vector

    return word_vectors


def filter_embeddings(trained_embedding, embedding_dim, vocabulary, zero_init_indices=0,
                      rand_init_indices=1):
    """Build word embeddings matrix from pre-trained-embeddings

    Args:
        trained_embedding: dict(str, np.array), a mapping of words to embeddings
        embedding_dim: int, dimensionality of embedding
        vocabulary: dict. a mapping of words to indices
        zero_init_indices: int or a list, the indices which use zero-initialization. These indices
                           usually represent padding token.
        rand_init_indices: int or a list, the indices which use randomly-initialization.These
                           indices usually represent other special tokens, such as "unk" token.

    Returns: np.array, a word embedding matrix.

    """
    emb = np.zeros(shape=(len(vocabulary), embedding_dim), dtype='float32')
    nb_unk = 0
    for w, i in vocabulary.items():
        if w not in trained_embedding:
            nb_unk += 1
            emb[i, :] = np.random.normal(0, 0.05, embedding_dim)
        else:
            emb[i, :] = trained_embedding[w]

    if isinstance(zero_init_indices, int):
        zero_init_indices = [zero_init_indices]
    if isinstance(rand_init_indices, int):
        rand_init_indices = [rand_init_indices]
    for idx in zero_init_indices:
        emb[idx] = np.zeros(embedding_dim)
    for idx in rand_init_indices:
        emb[idx] = np.random.normal(0, 0.05, embedding_dim)

    logging.info('Embedding matrix created, shaped: {}, not found tokens: {}'.format(emb.shape,
                                                                                     nb_unk))
    return emb


def load_pre_trained(load_filename, embedding_dim=None, vocabulary=None, zero_init_indices=0,
                     rand_init_indices=1):
    """Load pre-trained embedding and fit into vocabulary if provided

    Args:
        load_filename: str, pre-trained embedding file, in word2vec-like format or glove-like format
        embedding_dim: int, dimensionality of embeddings in the embedding file, must be provided when
                       the file is in glove-like format.
        vocabulary: dict. a mapping of words to indices
        zero_init_indices: int or a list, the indices which use zero-initialization. These indices
                           usually represent padding token.
        rand_init_indices: int or a list, the indices which use randomly-initialization.These
                           indices usually represent other special tokens, such as "unk" token.
    Returns: If vocabulary is None: dict(str, np.array), a mapping of words to embeddings.
             Otherwise: np.array, a word embedding matrix.

    """
    word_vectors = {}
    try:
        # load word2vec-like format pre-trained embedding file
        model = KeyedVectors.load_word2vec_format(load_filename)
        weights = model.wv.vectors
        word_embed_dim = weights.shape[1]
        for k, v in model.wv.vocab.items():
            word_vectors[k] = weights[v.index, :]
    except ValueError:
        # load glove-like format pre-trained embedding file
        if embedding_dim is None:
            raise ValueError('`embedding_dim` must be provided when pre-trained embedding file is'
                             'in glove-like format!')
        word_vectors = load_glove_format(load_filename, embedding_dim)
        word_embed_dim = embedding_dim

    if vocabulary is not None:
        emb = filter_embeddings(word_vectors, word_embed_dim, vocabulary, zero_init_indices,
                                rand_init_indices)
        return emb
    else:
        logging.info('Loading Embedding from: {}, shaped: {}'.format(load_filename,
                                                                     (len(word_vectors),
                                                                      word_embed_dim)))
        return word_vectors


def train_w2v(corpus, vocabulary, zero_init_indices=0, rand_init_indices=1, embedding_dim=300):
    """Use word2vec to train on corpus to obtain embedding

    Args:
        corpus: list of tokenized texts, corpus to train on
        vocabulary: dict, a mapping of words to indices
        zero_init_indices: int or a list, the indices which use zero-initialization. These indices
                           usually represent padding token.
        rand_init_indices: int or a list, the indices which use randomly-initialization.These
                           indices usually represent other special tokens, such as "unk" token.
        embedding_dim: int, dimensionality of embedding

    Returns: np.array, a word embedding matrix.

    """
    model = Word2Vec(corpus, size=embedding_dim, min_count=1, window=5, sg=1, iter=10)
    weights = model.wv.vectors
    d = dict([(k, v.index) for k, v in model.wv.vocab.items()])
    word_vectors = dict((w, weights[d[w], :]) for w in d)
    emb = filter_embeddings(word_vectors, embedding_dim, vocabulary, zero_init_indices,
                            rand_init_indices)
    return emb


def train_fasttext(corpus, vocabulary, zero_init_indices=0, rand_init_indices=1, embedding_dim=300):
    """Use fasttext to train on corpus to obtain embedding

        Args:
            corpus: list of tokenized texts, corpus to train on
            vocabulary: dict, a mapping of words to indices
            zero_init_indices: int or a list, the indices which use zero-initialization. These
                               indices usually represent padding token.
            rand_init_indices: int or a list, the indices which use randomly-initialization.These
                               indices usually represent other special tokens, such as "unk" token.
            embedding_dim: int, dimensionality of embedding

        Returns: np.array, a word embedding matrix.

        """
    model = FastText(size=embedding_dim, min_count=1, window=5, sg=1, word_ngrams=1)
    model.build_vocab(sentences=corpus)
    model.train(sentences=corpus, total_examples=len(corpus), epochs=10)

    emb = np.zeros(shape=(len(vocabulary), embedding_dim), dtype='float32')

    for w, i in vocabulary.items():
        emb[i, :] = model.wv[w]  # note that oov words can still have word vectors

    if isinstance(zero_init_indices, int):
        zero_init_indices = [zero_init_indices]
    if isinstance(rand_init_indices, int):
        rand_init_indices = [rand_init_indices]
    for idx in zero_init_indices:
        emb[idx] = np.zeros(embedding_dim)
    for idx in rand_init_indices:
        emb[idx] = np.random.normal(0, 0.05, embedding_dim)

    return emb



# -*- coding: utf-8 -*-

import tensorflow as tf
from keras.layers import *
from keras_contrib.layers import CRF
from keras_contrib.losses import crf_loss
from keras_contrib.metrics import crf_accuracy
from keras.models import Model
from keras.optimizers import Adam

from fancy_nlp.models.ner.base_ner_model import BaseNERModel


class BiLSTMNER(BaseNERModel):
    """Bidirectional LSTM model for NER.
       1. Support using CuDNNLSTM for acceleration when gpu is available. You will have to install
          a gpu version of tensorflow to do so. Note that model using CuDNNLSTM can only be deployed
          to machines with GPU for practical use.
       2. Support using CRF layer.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 fc_dim=100,
                 activation='tanh',
                 use_crf=True,
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_fim = fc_dim
        self.activation = activation
        self.use_crf = use_crf
        self.optimizer = optimizer
        super(BiLSTMNER, self).__init__(use_char, char_embeddings, char_vocab_size, char_embed_dim,
                                        char_embed_trainable, use_bert, bert_config_file,
                                        bert_checkpoint_file, bert_trainable, use_word,
                                        word_embeddings, word_vocab_size, word_embed_dim,
                                        word_embed_trainable, max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        if tf.test.is_gpu_available(cuda_only=True):
            input_encode = Bidirectional(CuDNNLSTM(self.rnn_units, return_sequences=True))(input_embed)
        else:
            input_encode = Bidirectional(LSTM(self.rnn_units, return_sequences=True))(input_embed)
        input_encode = TimeDistributed(Dense(self.fc_fim, activation=self.activation))(input_encode)

        if self.use_crf:
            crf = CRF(units=self.num_class, learn_mode='marginal')
            ner_tag = crf(input_encode)
            ner_loss = crf_loss
            ner_metrics = crf_accuracy
        else:
            ner_tag = TimeDistributed(Dense(self.num_class, activation='softmax'))(input_encode)
            ner_loss = 'categorical_crossentropy'
            ner_metrics = 'accuracy'
        ner_model = Model(model_inputs if len(model_inputs) > 1 else model_inputs[0], ner_tag)
        ner_model.compile(optimizer=self.optimizer, loss=ner_loss, metrics=[ner_metrics])
        return ner_model


class BiGRUNER(BaseNERModel):
    """Bidirectional GRU model for NER.
       1. Support using CuDNNGRU for acceleration when gpu is available. You will have to install
          a gpu version of tensorflow to do so. Note that model using CuDNNGRU can only be deployed
          to machines with GPU for practical use.
       2. Support using CRF layer.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 fc_dim=100,
                 activation='tanh',
                 use_crf=True,
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.fc_fim = fc_dim
        self.activation = activation
        self.use_crf = use_crf
        self.optimizer = optimizer
        super(BiGRUNER, self).__init__(use_char, char_embeddings, char_vocab_size, char_embed_dim,
                                       char_embed_trainable, use_bert, bert_config_file,
                                       bert_checkpoint_file, bert_trainable, use_word,
                                       word_embeddings, word_vocab_size, word_embed_dim,
                                       word_embed_trainable, max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        if tf.test.is_gpu_available(cuda_only=True):
            input_encode = Bidirectional(CuDNNGRU(self.rnn_units, return_sequences=True))(input_embed)
        else:
            input_encode = Bidirectional(GRU(self.rnn_units, return_sequences=True))(input_embed)
        input_encode = TimeDistributed(Dense(self.fc_fim, activation=self.activation))(input_encode)

        if self.use_crf:
            crf = CRF(units=self.num_class, learn_mode='marginal')
            ner_tag = crf(input_encode)
            ner_loss = crf_loss
            ner_metrics = crf_accuracy
        else:
            ner_tag = TimeDistributed(Dense(self.num_class, activation='softmax'))(input_encode)
            ner_loss = 'categorical_crossentropy'
            ner_metrics = 'accuracy'
        ner_model = Model(model_inputs if len(model_inputs) > 1 else model_inputs[0], ner_tag)
        ner_model.compile(optimizer=self.optimizer, loss=ner_loss, metrics=[ner_metrics])
        return ner_model


class BiLSTMCNNNER(BaseNERModel):
    """Bidirectional LSTM + CNN model for NER.
       1. Support using CuDNNLSTM for acceleration when gpu is available. You will have to install
          a gpu version of tensorflow to do so. Note that model using CuDNNLSTM can only be deployed
          to machines with GPU for practical use.
       2. Support using CRF layer.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 cnn_filters=300,
                 cnn_kernel_size=3,
                 fc_dim=100,
                 activation='tanh',
                 use_crf=True,
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.cnn_filters = cnn_filters
        self.cnn_kernel_size = cnn_kernel_size
        self.fc_fim = fc_dim
        self.activation = activation
        self.use_crf = use_crf
        self.optimizer = optimizer
        super(BiLSTMCNNNER, self).__init__(use_char, char_embeddings, char_vocab_size,
                                           char_embed_dim, char_embed_trainable,
                                           use_bert, bert_config_file, bert_checkpoint_file,
                                           bert_trainable, use_word, word_embeddings,
                                           word_vocab_size, word_embed_dim, word_embed_trainable,
                                           max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        if tf.test.is_gpu_available(cuda_only=True):
            input_encode = Bidirectional(CuDNNLSTM(self.rnn_units,
                                                   return_sequences=True))(input_embed)
        else:
            input_encode = Bidirectional(LSTM(self.rnn_units, return_sequences=True))(input_embed)
        input_encode = Conv1D(filters=self.cnn_filters, kernel_size=self.cnn_kernel_size,
                              padding='same', activation='relu')(input_encode)
        input_encode = TimeDistributed(Dense(self.fc_fim, activation=self.activation))(input_encode)

        if self.use_crf:
            crf = CRF(units=self.num_class, learn_mode='marginal')
            ner_tag = crf(input_encode)
            ner_loss = crf_loss
            ner_metrics = crf_accuracy
        else:
            ner_tag = TimeDistributed(Dense(self.num_class, activation='softmax'))(input_encode)
            ner_loss = 'categorical_crossentropy'
            ner_metrics = 'accuracy'
        ner_model = Model(model_inputs if len(model_inputs) > 1 else model_inputs[0], ner_tag)
        ner_model.compile(optimizer=self.optimizer, loss=ner_loss, metrics=[ner_metrics])
        return ner_model


class BiGRUCNNNER(BaseNERModel):
    """Bidirectional GRU + CNN model for NER.
       1. Support using CuDNNGRU for acceleration when gpu is available. You will have to install
          a gpu version of tensorflow to do so. Note that model using CuDNNGRU can only be deployed
          to machines with GPU for practical use.
       2. Support using CRF layer.
    """
    def __init__(self,
                 num_class,
                 use_char=True,
                 char_embeddings=None,
                 char_vocab_size=-1,
                 char_embed_dim=-1,
                 char_embed_trainable=False,
                 use_bert=False,
                 bert_config_file=None,
                 bert_checkpoint_file=None,
                 bert_trainable=False,
                 use_word=False,
                 word_embeddings=None,
                 word_vocab_size=-1,
                 word_embed_dim=-1,
                 word_embed_trainable=False,
                 max_len=None,
                 dropout=0.2,
                 rnn_units=150,
                 cnn_filters=300,
                 cnn_kernel_size=3,
                 fc_dim=100,
                 activation='tanh',
                 use_crf=True,
                 optimizer='adam'):
        self.num_class = num_class
        self.rnn_units = rnn_units
        self.cnn_filters = cnn_filters
        self.cnn_kernel_size = cnn_kernel_size
        self.fc_fim = fc_dim
        self.activation = activation
        self.use_crf = use_crf
        self.optimizer = optimizer
        super(BiGRUCNNNER, self).__init__(use_char, char_embeddings, char_vocab_size,
                                          char_embed_dim, char_embed_trainable,
                                          use_bert, bert_config_file, bert_checkpoint_file,
                                          bert_trainable, use_word, word_embeddings,
                                          word_vocab_size, word_embed_dim, word_embed_trainable,
                                          max_len, dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        if tf.test.is_gpu_available(cuda_only=True):
            input_encode = Bidirectional(CuDNNGRU(self.rnn_units,
                                                  return_sequences=True))(input_embed)
        else:
            input_encode = Bidirectional(GRU(self.rnn_units, return_sequences=True))(input_embed)
        input_encode = Conv1D(filters=self.cnn_filters, kernel_size=self.cnn_kernel_size,
                              padding='same', activation='relu')(input_encode)
        input_encode = TimeDistributed(Dense(self.fc_fim, activation=self.activation))(input_encode)

        if self.use_crf:
            crf = CRF(units=self.num_class, learn_mode='marginal')
            ner_tag = crf(input_encode)
            ner_loss = crf_loss
            ner_metrics = crf_accuracy
        else:
            ner_tag = TimeDistributed(Dense(self.num_class, activation='softmax'))(input_encode)
            ner_loss = 'categorical_crossentropy'
            ner_metrics = 'accuracy'
        ner_model = Model(model_inputs if len(model_inputs) > 1 else model_inputs[0], ner_tag)
        ner_model.compile(optimizer=self.optimizer, loss=ner_loss, metrics=[ner_metrics])
        return ner_model


class BertNER(BaseNERModel):
    """Bert model for NER. Support using CRF layer.
    We suggest you to train bert on machines with GPU cause it will be very slow to be trained with
    cpu. You will have to re-install a gpu version of tensorflow to do so.
    """

    def __init__(self,
                 num_class,
                 bert_config_file,
                 bert_checkpoint_file,
                 bert_trainable,
                 max_len,
                 dropout=0.2,
                 fc_dim=100,
                 activation='tanh',
                 use_crf=True,
                 optimizer=Adam(lr=1e-5)):  # use a small learning rate for bert
        self.num_class = num_class
        self.fc_fim = fc_dim
        self.activation = activation
        self.use_crf = use_crf
        self.optimizer = optimizer
        super(BertNER, self).__init__(use_char=False, use_bert=True,
                                      bert_config_file=bert_config_file,
                                      bert_checkpoint_file=bert_checkpoint_file,
                                      bert_trainable=bert_trainable, use_word=False,
                                      max_len=max_len, dropout=dropout)

    def build_model(self):
        model_inputs, input_embed = self.build_input()
        input_encode = TimeDistributed(Dense(self.fc_fim, activation=self.activation))(input_embed)

        if self.use_crf:
            crf = CRF(units=self.num_class, learn_mode='marginal')
            ner_tag = crf(input_encode)
            ner_loss = crf_loss
            ner_metrics = crf_accuracy
        else:
            ner_tag = TimeDistributed(Dense(self.num_class, activation='softmax'))(input_encode)
            ner_loss = 'categorical_crossentropy'
            ner_metrics = 'accuracy'
        ner_model = Model(model_inputs if len(model_inputs) > 1 else model_inputs[0], ner_tag)
        ner_model.compile(optimizer=self.optimizer, loss=ner_loss, metrics=[ner_metrics])
        return ner_model



# -*- coding: utf-8 -*-

from keras import backend as K
from keras.engine.topology import Layer


class FullMatching(Layer):
    """
       Full Matching strategy, each contextual embedding is compared with the average
       representation of the other sentence.
    """
    def __init__(self, perspective_num=10, **kwargs):
        self.perspective_num = perspective_num
        self.kernel = None
        super(FullMatching, self).__init__(**kwargs)

    def build(self, input_shape):
        self.dim = input_shape[0][-1]
        self.max_len = input_shape[0][1]
        self.kernel = self.add_weight(name='kernel', shape=(self.perspective_num, self.dim),
                                      initializer='glorot_uniform')
        super(FullMatching, self).build(input_shape)

    def call(self, inputs, **kwargs):
        sent1 = inputs[0]
        sent2 = inputs[1]

        v1 = K.expand_dims(sent1, -2) * self.kernel
        v2 = self.kernel * K.expand_dims(sent2, 1)
        v2 = K.expand_dims(v2, 1)
        v1 = K.l2_normalize(v1, axis=-1)
        v2 = K.l2_normalize(v2, axis=-1)
        matching = K.sum(v1 * v2, axis=-1)
        return matching

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], input_shape[0][1], self.perspective_num

    def get_config(self):
        config = {'perspective_num': self.perspective_num}
        base_config = super(FullMatching, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class MaxPoolingMatching(Layer):
    """
       MaxPooling Matching strategy, each contextual embedding is compared with every
       contextual embeddings of the other sentence, and only the maximum value of each
       dimension is retained.
    """
    def __init__(self, perspective_num=10, **kwargs):
        self.perspective_num = perspective_num
        self.kernel = None
        super(MaxPoolingMatching, self).__init__(**kwargs)

    def build(self, input_shape):
        self.dim = input_shape[0][-1]
        self.max_len = input_shape[0][1]
        self.kernel = self.add_weight(name='kernel', shape=(self.perspective_num, self.dim),
                                      initializer='glorot_uniform')
        super(MaxPoolingMatching, self).build(input_shape)

    def call(self, inputs, **kwargs):
        sent1 = inputs[0]
        sent2 = inputs[1]

        v1 = K.expand_dims(sent1, -2) * self.kernel
        v2 = K.expand_dims(sent2, -2) * self.kernel
        v1 = K.l2_normalize(v1, axis=-1)
        v2 = K.l2_normalize(v2, axis=-1)
        matching = K.max(K.sum(K.expand_dims(v1, 2) * K.expand_dims(v2, 1), axis=-1), axis=-2)
        return matching

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], input_shape[0][1], self.perspective_num

    def get_config(self):
        config = {'perspective_num': self.perspective_num}
        base_config = super(MaxPoolingMatching, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AttentiveMatching(Layer):
    """
       Attentive Matching strategy, each contextual embedding is compared with its attentive
       weighted representation of the other sentence.
    """
    def __init__(self, perspective_num=10, **kwargs):
        self.perspective_num = perspective_num
        self.kernel = None
        super(AttentiveMatching, self).__init__(**kwargs)

    def build(self, input_shape):
        self.dim = input_shape[0][-1]
        self.max_len = input_shape[0][1]
        self.kernel = self.add_weight(name='kernel', shape=(self.perspective_num, self.dim),
                                      initializer='glorot_uniform')
        super(AttentiveMatching, self).build(input_shape)

    def call(self, inputs, **kwargs):
        sent1 = inputs[0]
        sent2 = inputs[1]

        v1 = K.expand_dims(sent1, -2) * self.kernel
        v2 = K.expand_dims(sent2, -2) * self.kernel
        v1 = K.l2_normalize(v1, axis=-1)
        v2 = K.l2_normalize(v2, axis=-1)
        matching = K.sum(v1 * v2, axis=-1)
        return matching

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], input_shape[0][1], self.perspective_num

    def get_config(self):
        config = {'perspective_num': self.perspective_num}
        base_config = super(AttentiveMatching, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class MaxAttentiveMatching(Layer):
    """
       MaxAttentive Matching strategy, each contextual embedding picks the contextual
       embedding of the other sentence with the highest cosine similarity as the
       attentive vector.
    """
    def __init__(self, perspective_num=10, **kwargs):
        self.perspective_num = perspective_num
        self.kernel = None
        super(MaxAttentiveMatching, self).__init__(**kwargs)

    def build(self, input_shape):
        self.dim = input_shape[0][-1]
        self.max_len = input_shape[0][1]
        self.kernel = self.add_weight(name='kernel', shape=(self.perspective_num, self.dim),
                                      initializer='glorot_uniform')
        super(MaxAttentiveMatching, self).build(input_shape)

    def call(self, inputs, **kwargs):
        sent1 = inputs[0]
        sent2 = inputs[1]

        v1 = K.expand_dims(sent1, -2) * self.kernel
        v2 = K.expand_dims(sent2, -2) * self.kernel
        v1 = K.l2_normalize(v1, axis=-1)
        v2 = K.l2_normalize(v2, axis=-1)
        matching = K.sum(v1 * v2, axis=-1)
        return matching

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], input_shape[0][1], self.perspective_num

    def get_config(self):
        config = {'perspective_num': self.perspective_num}
        base_config = super(MaxAttentiveMatching, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from .ner_preprocessor import NERPreprocessor
from .text_classification_preprocessor import TextClassificationPreprocessor
from .spm_preprocessor import SPMPreprocessor



# -*- coding: utf-8 -*-

from keras.callbacks import Callback
from seqeval import metrics
from sklearn.metrics import f1_score, precision_score, recall_score, classification_report
import numpy as np


class NERMetric(Callback):
    """
    callback for evaluating ner model
    """
    def __init__(self, preprocessor, valid_data, valid_labels):
        """
        Args:
            preprocessor: `NERPreprocessor` instance to help prepare input for ner model
            valid_data: list of tokenized texts (, like ``[['我', '是', '中', '国', '人']]``
            valid_labels: list of list of str, the corresponding label strings
        """
        self.preprocessor = preprocessor
        self.valid_data = valid_data
        self.valid_labels = valid_labels
        self.valid_features, self.valid_y = self.preprocessor.prepare_input(valid_data,
                                                                            valid_labels)
        super(NERMetric, self).__init__()

    def get_lengths(self, pred_probs):
        return [min(len(valid_label), pred_prob.shape[0])
                for valid_label, pred_prob in zip(self.valid_labels, pred_probs)]

    def on_epoch_end(self, epoch, logs=None):
        pred_probs = self.model.predict(self.valid_features)
        if self.preprocessor.use_bert:
            pred_probs = pred_probs[:, 1:-1, :]     # remove <CLS> and <SEQ>
        y_pred = self.preprocessor.label_decode(pred_probs, self.get_lengths(pred_probs))

        r = metrics.recall_score(self.valid_labels, y_pred)
        p = metrics.precision_score(self.valid_labels, y_pred)
        f1 = metrics.f1_score(self.valid_labels, y_pred)

        logs['val_r'] = r
        logs['val_p'] = p
        logs['val_f1'] = f1
        print('Epoch {}: val_r: {}, val_p: {}, val_f1: {}'.format(epoch+1, r, p, f1))
        print(metrics.classification_report(self.valid_labels, y_pred))


class TextClassificationMetric(Callback):
    """
    callback for evaluating text classification model
    """
    def __init__(self, preprocessor, valid_data, valid_labels):
        """
        Args:
            preprocessor: `TextClassificationPreprocessor` instance to help prepare input for
            text classification model
            valid_data: list of tokenized texts (, like ``[['我', '是', '中', '国', '人']]``
            valid_labels: list of str, the corresponding label strings
        """
        self.preprocessor = preprocessor
        self.valid_data = valid_data
        self.valid_labels = valid_labels
        self.valid_features, self.valid_y = self.preprocessor.prepare_input(valid_data,
                                                                            valid_labels)
        super(TextClassificationMetric, self).__init__()

    def on_epoch_end(self, epoch, logs=None):
        pred_probs = self.model.predict(self.valid_features)
        y_pred = self.preprocessor.label_decode(pred_probs)

        r = recall_score(self.valid_labels, y_pred, average='macro')
        p = precision_score(self.valid_labels, y_pred, average='macro')
        f1 = f1_score(self.valid_labels, y_pred, average='macro')

        logs['val_r'] = r
        logs['val_p'] = p
        logs['val_f1'] = f1
        print('Epoch {}: val_r: {}, val_p: {}, val_f1: {}'.format(epoch + 1, r, p, f1))
        print(classification_report(self.valid_labels, y_pred))


class SPMMetric(Callback):
    """
    callback for evaluating spm model
    """
    def __init__(self, preprocessor, valid_data, valid_labels):
        """
        Args:
            preprocessor: `SPMPreprocessor` instance to help prepare input for spm model
            valid_data: list of text pairs (, like ``[['我是中国人', ...], ['我爱中国', ...]]``
            valid_labels: list of str, the corresponding label strings
        """
        self.preprocessor = preprocessor
        self.valid_data = valid_data
        self.valid_labels = valid_labels
        self.valid_features, self.valid_y = self.preprocessor.prepare_input(valid_data,
                                                                            valid_labels)
        super(SPMMetric, self).__init__()

    def on_epoch_end(self, epoch, logs=None):
        pred_probs = self.model.predict(self.valid_features)
        y_pred = np.argmax(pred_probs, axis=-1)
        y_true = np.argmax(self.valid_y, axis=-1)

        r = recall_score(y_true, y_pred, average='macro')
        p = precision_score(y_true, y_pred, average='macro')
        f1 = f1_score(y_true, y_pred, average='macro')

        logs['val_r'] = r
        logs['val_p'] = p
        logs['val_f1'] = f1
        print('Epoch {}: val_r: {}, val_p: {}, val_f1: {}'.format(epoch, r, p, f1))
        print(classification_report(y_true, y_pred))



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from .data_loader import load_ner_data_and_labels, load_text_classification_data_and_labels, \
    load_spm_data_and_labels
from .embedding import load_pre_trained, train_w2v, train_fasttext
from .data_generator import NERGenerator, TextClassificationGenerator, SPMGenerator
from .save_load_model import load_keras_model, save_keras_model
from .other import pad_sequences_2d, get_len_from_corpus, get_custom_objects, ChineseBertTokenizer



# -*- coding: utf-8 -*-

from absl import logging
from keras.models import model_from_json
from keras.utils import get_file

from fancy_nlp.preprocessors import SPMPreprocessor
from fancy_nlp.models.spm import *
from fancy_nlp.trainers import SPMTrainer
from fancy_nlp.predictors import SPMPredictor
from fancy_nlp.utils import get_custom_objects
from fancy_nlp.config import CACHE_DIR, MODEL_STORAGE_PREFIX


class SPM(object):
    """SPM application"""

    def __init__(self, use_pretrained=True):
        self.preprocessor = None
        self.model = None
        self.trainer = None
        self.predictor = None

        if use_pretrained:
            self.load_pretrained_model()

    def fit(self,
            train_data,
            train_labels,
            valid_data=None,
            valid_labels=None,
            spm_model_type='siamese_cnn',
            use_word=True,
            external_word_dict=None,
            word_embed_type='word2vec',
            word_embed_dim=300,
            word_embed_trainable=True,
            use_char=False,
            char_embed_type='word2vec',
            char_embed_dim=300,
            char_embed_trainable=True,
            use_bert=False,
            bert_vocab_file=None,
            bert_config_file=None,
            bert_checkpoint_file=None,
            bert_trainable=False,
            bert_output_layer_num=1,
            label_dict_file=None,
            max_len=None,
            max_word_len=None,
            optimizer='adam',
            batch_size=32,
            epochs=50,
            callback_list=None,
            checkpoint_dir=None,
            model_name=None,
            load_swa_model=False,
            **kwargs):
        """Train spm model using provided data

        Args:
            train_data: list of untokenized text pairs for training,
                        like ``[['我是中国人', ...], ['我爱中国', ...]]``
            train_labels: labels string of train_data
            valid_data: list of untokenized text pairs for evaluation
            valid_labels: labels string of valid data
            spm_model_type: str, which spm model to use
            use_word: boolean, whether to use word embedding as input
            external_word_dict: external word dictionary
            word_embed_type: str, can be a pre-trained embedding filename or pre-trained embedding
                             methods (word2vec, glove, fastext)
            word_embed_dim: int, dimensionality of word embedding
            word_embed_trainable: boolean, whether to update word embedding during training
            use_char: boolean, whether to use char as input
            char_embed_type: similar as 'word_embed_type'
            char_embed_dim: similar as 'word_embed_dim'
            char_embed_trainable: similar as 'word_embed_trainable'
            use_bert: boolean, whether to use bert embedding as input
            bert_vocab_file: str, path to bert's vocabulary file
            bert_config_file: str, path to bert's configuration file
            bert_checkpoint_file: str, path to bert's checkpoint file
            bert_trainable: boolean, whether to update bert during training
            bert_output_layer_num: int, the number of bert layers
            use_bert_model: boolean, whether to use traditional bert model which combines two sentences
                            as one input
            label_dict_file: a file with two columns separated by tab, the first column is raw
                             label name, and the second column is the corresponding name which is
                             meaningful
            max_len: int, max sequence length. If None, we dynamically use the max length of one batch
                     as max_len. However, max_len must be provided when using bert as input.
            max_word_len: int max word length. If None, we dynamically use the max word length of one
                          batch as max_word_len.
            optimizer: str or instance of `keras.optimizers.Optimizer`, indicating the optimizer to
                       use during training
            batch_size: num of samples per gradient update
            epochs: num of epochs to train the model
            callback_list: list of str, each item indicates the callback to apply during training
                           Currently, we support using 'modelcheckpoint' for `ModelCheckpoint`
                           callback, 'earlystopping` for 'Earlystopping` callback, 'swa' for
                           'SWA' callback. We will automatically add `SPMMetric` callback when
                           valid_data and valid_labels are both provided.
            checkpoint_dir: str, directory to save spm model, must be provided when using
                            `ModelCheckpoint` or `SWA` callback.
            model_name: str, prefix of spm model's weights file must be provided when using
                        `ModelCheckpoint` or `SWA` callback.
                        For example, if checkpoint_dir is 'ckpt' and model_name is 'model', the
                        weights of spm model saved by `ModelCheckpoint` callback will be
                        'ckpt/model.hdf5' and by `SWA` callback will be 'ckpt/model_swa.hdf5'
            load_swa_model: boolean, whether to load swa model, only apply when using SWA Callback.
            **kwargs: other argument for building spm model, such as "rnn_units", "fc_dim" etc.
        """
        use_bert_model = True if spm_model_type == 'bert' else False

        self.preprocessor = SPMPreprocessor(train_data=train_data,
                                            train_labels=train_labels,
                                            use_word=use_word,
                                            use_char=use_char,
                                            use_bert=use_bert,
                                            use_bert_model=use_bert_model,
                                            external_word_dict=external_word_dict,
                                            bert_vocab_file=bert_vocab_file,
                                            char_embed_type=char_embed_type,
                                            char_embed_dim=char_embed_dim,
                                            word_embed_type=word_embed_type,
                                            word_embed_dim=word_embed_dim,
                                            label_dict_file=label_dict_file,
                                            max_len=max_len,
                                            max_word_len=max_word_len)

        self.model = self.get_spm_model(spm_model_type=spm_model_type,
                                        num_class=self.preprocessor.num_class,
                                        use_word=use_word,
                                        word_embeddings=self.preprocessor.word_embeddings,
                                        word_vocab_size=self.preprocessor.word_vocab_size,
                                        word_embed_dim=self.preprocessor.word_embed_dim,
                                        word_embed_trainable=word_embed_trainable,
                                        use_char=use_char,
                                        char_embeddings=self.preprocessor.char_embeddings,
                                        char_vocab_size=self.preprocessor.char_vocab_size,
                                        char_embed_dim=self.preprocessor.char_embed_dim,
                                        char_embed_trainable=char_embed_trainable,
                                        use_bert=use_bert,
                                        bert_config_file=bert_config_file,
                                        bert_checkpoint_file=bert_checkpoint_file,
                                        bert_trainable=bert_trainable,
                                        bert_output_layer_num=bert_output_layer_num,
                                        max_len=self.preprocessor.max_len,
                                        max_word_len=self.preprocessor.max_word_len,
                                        optimizer=optimizer,
                                        **kwargs)

        if 'swa' in callback_list:
            swa_model = self.get_spm_model(spm_model_type=spm_model_type,
                                           num_class=self.preprocessor.num_class,
                                           use_word=use_word,
                                           word_embeddings=self.preprocessor.word_embeddings,
                                           word_vocab_size=self.preprocessor.word_vocab_size,
                                           word_embed_dim=self.preprocessor.word_embed_dim,
                                           word_embed_trainable=word_embed_trainable,
                                           use_char=use_char,
                                           char_embeddings=self.preprocessor.char_embeddings,
                                           char_vocab_size=self.preprocessor.char_vocab_size,
                                           char_embed_dim=self.preprocessor.char_embed_dim,
                                           char_embed_trainable=char_embed_trainable,
                                           use_bert=use_bert,
                                           bert_config_file=bert_config_file,
                                           bert_checkpoint_file=bert_checkpoint_file,
                                           bert_trainable=bert_trainable,
                                           bert_output_layer_num=bert_output_layer_num,
                                           max_len=self.preprocessor.max_len,
                                           max_word_len=self.preprocessor.max_word_len,
                                           optimizer=optimizer,
                                           **kwargs)
        else:
            swa_model = None

        self.trainer = SPMTrainer(self.model, self.preprocessor)
        self.trainer.train_generator(train_data, train_labels, valid_data, valid_labels,
                                     batch_size, epochs, callback_list, checkpoint_dir, model_name,
                                     swa_model, load_swa_model)

        self.predictor = SPMPredictor(self.model, self.preprocessor)

        if valid_data is not None and valid_labels is not None:
            logging.info('Evaluating on validation data...')
            self.score(valid_data, valid_labels)

    def score(self, valid_data, valid_labels):
        """Return the f1 score of the model over validation data

        Args:
            valid_data: list of untokenized text pairs
            valid_labels: list of label strings

        Returns:

        """
        if self.trainer:
            return self.trainer.evaluate(valid_data, valid_labels)
        else:
            logging.fatal('Trainer is None! Call fit() or load() to get trainer.')

    def predict(self, test_text):
        """Return prediction of the model for test data

        Args:
            test_text: a pair of untokenized text

        Returns:

        """
        if self.predictor:
            return self.predictor.matching(test_text)
        else:
            logging.fatal('Predictor is None! Call fit() or load() to get predictor.')

    def predict_batch(self, test_texts):
        """Return predictions of the model for test data

        Args:
            test_texts: list of untokenized text pairs

        Returns:

        """
        if self.predictor:
            return self.predictor.matching_batch(test_texts)
        else:
            logging.fatal('Predictor is None! Call fit() or load() to get predictor.')

    def analyze(self, text):
        """Analyze text and return matching result with probability.

        Args:
            text: a pair of untokenized text
        Returns:

        """
        if self.predictor:
            return self.predictor.matching_with_prob(text)
        else:
            logging.fatal('Predictor is None! Call fit() or load() to get predictor.')

    def analyze_batch(self, texts):
        """Analyze text and return matching result with probability.

        Args:
            texts: list of untokenized text pairs
        Returns:

        """
        if self.predictor:
            return self.predictor.matching_with_prob_batch(texts)
        else:
            logging.fatal('Predictor is None! Call fit() or load() to get predictor.')

    def save(self, preprocessor_file, json_file, weights_file=None):
        """save spm application

        Args:
            preprocessor_file: path to save preprocessor
            json_file: path to save model architecture
            weights_file: path to save model weights, can be None. When we use `ModelCheckpoint`
                          or `SWA` callback, model's weights will be saved to disk after training.
                          In that case, we don't need to save it again. We usually set weights_file
                          to be None.
        """
        self.preprocessor.save(preprocessor_file)
        logging.info('Save preprocessor to {}'.format(preprocessor_file))

        model_json = self.model.to_json()
        with open(json_file, 'w') as writer:
            writer.write(model_json)
        logging.info('Save model architecture to {}'.format(json_file))

        if weights_file:
            self.model.save_weights(weights_file)
            logging.info('Save model weights to {}'.format(weights_file))

    def load(self, preprocessor_file, json_file, weights_file, custom_objects=None):
        """load spm application

        Args:
            preprocessor_file: path to load preprocessor
            json_file: path to load model architecture
            weights_file: path to load model weights
            custom_objects: Optional dictionary mapping names (strings) to custom classes or
                            functions to be considered during deserialization. Must provided when
                            using custom layer.

        """
        self.preprocessor = SPMPreprocessor.load(preprocessor_file)
        logging.info('Load preprocessor from {}'.format(preprocessor_file))

        custom_objects = custom_objects or {}
        custom_objects.update(get_custom_objects())
        with open(json_file, 'r') as reader:
            self.model = model_from_json(reader.read(), custom_objects=custom_objects)
        logging.info('Load model architecture from {}'.format(json_file))

        self.model.load_weights(weights_file)
        logging.info('Load model weight from {}'.format(weights_file))

        self.trainer = SPMTrainer(self.model, self.preprocessor)
        self.predictor = SPMPredictor(self.model, self.preprocessor)

    @staticmethod
    def get_spm_model(spm_model_type, num_class, use_word, word_embeddings, word_vocab_size,
                      word_embed_dim, word_embed_trainable, use_char, char_embeddings,
                      char_vocab_size, char_embed_dim, char_embed_trainable, use_bert,
                      bert_config_file, bert_checkpoint_file, bert_trainable,
                      bert_output_layer_num, max_len, max_word_len, optimizer, **kwargs):
        spm_model_all = {'siamese_cnn': SiameseCNN,
                         'siamese_bilstm': SiameseBiLSTM,
                         'siamese_bigru': SiameseBiGRU,
                         'esim': ESIM,
                         'bimpm': BiMPM,
                         'bert': BertSPM}
        if spm_model_type in spm_model_all:
            spm_model = spm_model_all[spm_model_type](
                num_class=num_class,
                use_word=use_word,
                word_embeddings=word_embeddings,
                word_vocab_size=word_vocab_size,
                word_embed_dim=word_embed_dim,
                word_embed_trainable=word_embed_trainable,
                use_char=use_char,
                char_embeddings=char_embeddings,
                char_vocab_size=char_vocab_size,
                char_embed_dim=char_embed_dim,
                char_embed_trainable=char_embed_trainable,
                use_bert=use_bert,
                bert_config_file=bert_config_file,
                bert_checkpoint_file=bert_checkpoint_file,
                bert_trainable=bert_trainable,
                bert_output_layer_num=bert_output_layer_num,
                max_len=max_len,
                max_word_len=max_word_len,
                optimizer=optimizer,
                **kwargs)

        else:
            raise ValueError('`spm_model_type` not understood: {}'.format(spm_model_type))

        return spm_model.build_model()

    def load_pretrained_model(self):
        cache_subdir = 'pretrained_models'

        preprocessor_file = get_file(
            fname='webank_spm_siamese_cnn_word_preprocessor.pkl',
            origin=MODEL_STORAGE_PREFIX + 'webank_spm_siamese_cnn_word_preprocessor.pkl',
            cache_subdir=cache_subdir, cache_dir=CACHE_DIR)
        json_file = get_file(
            fname='webank_spm_siamese_cnn_word.json',
            origin=MODEL_STORAGE_PREFIX + 'webank_spm_siamese_cnn_word.json',
            cache_subdir=cache_subdir, cache_dir=CACHE_DIR)
        weights_file = get_file(
            fname='webank_spm_siamese_cnn_word.hdf5',
            origin=MODEL_STORAGE_PREFIX + 'webank_spm_siamese_cnn_word.hdf5',
            cache_subdir=cache_subdir, cache_dir=CACHE_DIR)

        self.load(preprocessor_file, json_file, weights_file)





# -*- coding: utf-8 -*-

from absl import logging
from keras.callbacks import *
from seqeval import metrics

from fancy_nlp.utils import NERGenerator
from fancy_nlp.callbacks import NERMetric
from fancy_nlp.callbacks import SWA


class NERTrainer(object):
    def __init__(self, model, preprocessor):
        """

        Args:
            model: instance of keras Model
            preprocessor: instance of NERPreporcessor
        """
        self.model = model
        self.preprocessor = preprocessor

    def train(self, train_data, train_labels, valid_data=None, valid_labels=None, batch_size=32,
              epochs=50, callback_list=None, checkpoint_dir=None, model_name=None, swa_model=None,
              load_swa_model=False):
        callbacks = self.prepare_callback(callback_list, valid_data, valid_labels, checkpoint_dir,
                                          model_name, swa_model)

        train_features, train_y = self.preprocessor.prepare_input(train_data, train_labels)
        if valid_data is not None and valid_labels is not None:
            valid_features, valid_y = self.preprocessor.prepare_input(valid_data, valid_labels)
            validation_data = (valid_features, valid_y)
        else:
            validation_data = None

        logging.info('Training start...')
        self.model.fit(x=train_features, y=train_y, batch_size=batch_size, epochs=epochs,
                       validation_data=validation_data, callbacks=callbacks)
        logging.info('Training end...')

        if load_swa_model and callback_list is not None and 'swa' in callback_list:
            logging.info('Loading swa model after using SWA callback')
            self.load_model_weights(os.path.join(checkpoint_dir, f'{model_name}_swa.hdf5'))

        elif callback_list is not None and 'modelcheckpoint' in callback_list:
            logging.info('Loading best model after using ModelCheckpoint callback...')
            self.load_model_weights(os.path.join(checkpoint_dir, f'{model_name}.hdf5'))

    def train_generator(self, train_data, train_labels, valid_data=None, valid_labels=None,
                        batch_size=32, epochs=50, callback_list=None, checkpoint_dir=None,
                        model_name=None, swa_model=None, load_swa_model=False):
        callbacks = self.prepare_callback(callback_list, valid_data, valid_labels, checkpoint_dir,
                                          model_name, swa_model)

        train_generator = NERGenerator(self.preprocessor, train_data, train_labels, batch_size)

        if valid_data and valid_labels:
            valid_generator = NERGenerator(self.preprocessor, valid_data, valid_labels,
                                           batch_size)
        else:
            valid_generator = None

        print('Training start...')
        self.model.fit_generator(generator=train_generator, epochs=epochs,
                                 validation_data=valid_generator, callbacks=callbacks)
        print('Training end...')

        if load_swa_model and callback_list is not None and 'swa' in callback_list:
            logging.info('Loading swa model after using SWA callback')
            self.load_model_weights(os.path.join(checkpoint_dir, f'{model_name}_swa.hdf5'))

        elif callback_list is not None and 'modelcheckpoint' in callback_list:
            logging.info('Loading best model after using ModelCheckpoint callback...')
            self.load_model_weights(os.path.join(checkpoint_dir, f'{model_name}.hdf5'))

    def prepare_callback(self, callback_list, valid_data=None, valid_labels=None,
                         checkpoint_dir=None, model_name=None, swa_model=None):
        """

        Args:
            callback_list: list of str, each item indicate the callback to apply during training.
                       For example, 'earlystopping' means using 'EarlyStopping' callback.
            valid_data: list of tokenized (in char level) texts for evaluation
            valid_labels: labels string of valid data
            checkpoint_dir: str, directory to save ner model, must be provided when using
                            `ModelCheckpoint` or `SWA` callback.
            model_name: str, prefix of ner model's weights filem must be provided when using
                        `ModelCheckpoint` or `SWA` callback.
                        For example, if checkpoint_dir is 'ckpt' and model_name is 'model', the
                        weights of ner model saved by `ModelCheckpoint` callback will be
                        'ckpt/model.hdf5' and by `SWA` callback will be 'ckpt/model_swa.hdf5'

        Returns: a list of `keras.callbacks.Callback` instances

        """
        assert not isinstance(callback_list, str)
        callback_list = callback_list or []
        callbacks = []
        if valid_data is not None and valid_labels is not None:
            callbacks.append(NERMetric(self.preprocessor, valid_data, valid_labels))
            add_metric = True
        else:
            add_metric = False

        if 'modelcheckpoint' in callback_list:
            if not add_metric:
                logging.warning('Using `ModelCheckpoint` with validation data not provided is not '
                                'Recommended! We will use `loss` (of training data) as monitor.')

            assert checkpoint_dir is not None, \
                '"checkpoint_dir" must must be provided when using "ModelCheckpoint" callback'
            assert model_name is not None, \
                '"model_name" must must be provided when using "ModelCheckpoint" callback'
            callbacks.append(ModelCheckpoint(filepath=os.path.join(checkpoint_dir,
                                                                   f'{model_name}.hdf5'),
                                             monitor='val_f1' if add_metric else 'loss',
                                             save_best_only=True,
                                             save_weights_only=True,
                                             mode='max' if add_metric else 'min',
                                             verbose=1))
            logging.info('ModelCheckpoint Callback added')

        if 'earlystopping' in callback_list:
            if not add_metric:
                logging.warning('Using `Earlystopping` with validation data not provided is not '
                                'Recommended! We will use `loss` (of training data) as monitor.')
            callbacks.append(EarlyStopping(monitor='val_f1' if add_metric else 'loss',
                                           mode='max' if add_metric else 'min',
                                           patience=5,
                                           verbose=1))
            logging.info('Earlystopping Callback added')

        if 'swa' in callback_list:
            assert checkpoint_dir is not None, \
                '"checkpoint_dir" must must be provided when using "SWA" callback'
            assert model_name is not None, \
                '"model_name" must must be provided when using "SWA" callback'
            callbacks.append(SWA(swa_model=swa_model, checkpoint_dir=checkpoint_dir,
                                 model_name=model_name, swa_start=5))
            logging.info('SWA Callback added')

        return callbacks

    def load_model_weights(self, weights_file):
        self.model.load_weights(weights_file)

    def evaluate(self, data, labels):
        """Evaluate the performance of ner model.

        Args:
            data: list of tokenized texts (, like ``[['我', '是', '中', '国', '人']]``
            labels: list of list of str, the corresponding label strings

        """
        features, y = self.preprocessor.prepare_input(data, labels)
        pred_probs = self.model.predict(features)
        if self.preprocessor.use_bert:
            pred_probs = pred_probs[:, 1:-1, :]     # remove <CLS> and <SEQ>

        lengths = [min(len(label), pred_prob.shape[0])
                   for label, pred_prob in zip(labels, pred_probs)]
        y_pred = self.preprocessor.label_decode(pred_probs, lengths)

        r = metrics.recall_score(labels, y_pred)
        p = metrics.precision_score(labels, y_pred)
        f1 = metrics.f1_score(labels, y_pred)

        print('Recall: {}, Precision: {}, F1: {}'.format(r, p, f1))
        print(metrics.classification_report(labels, y_pred))
        return f1



# -*- coding: utf-8 -*-

import pickle
import codecs

import numpy as np
import jieba
from itertools import chain
from absl import logging
from keras.utils.np_utils import to_categorical

from fancy_nlp.preprocessors.preprocessor import Preprocessor
from fancy_nlp.utils import pad_sequences_2d, get_len_from_corpus, ChineseBertTokenizer


class SPMPreprocessor(Preprocessor):
    """SPM preprocessor.
    """
    def __init__(self, train_data, train_labels, min_count=2, use_word=False, use_char=True,
                 use_bert=False, use_bert_model=False, external_word_dict=None, bert_vocab_file=None,
                 word_embed_type=None, word_embed_dim=300, char_embed_type=None, char_embed_dim=300,
                 label_dict_file=None, max_len=None, max_word_len=None,
                 padding_mode='post', truncating_mode='post'):
        """

        Args:
            train_data: a list of untokenized text pairs
            train_labels: list of str, train_data's labels
            min_count: int, token of which frequency is lower than min_count will be ignored
            use_word: whether to use word embedding as input
            use_char：whether to use char embedding as input
            use_bert: whether to use bert embedding as input
            use_bert_model: boolean, whether to use traditional bert model which combines two
                            sentences as one input
            word_embed_type: str, can be a pre-trained embedding filename or pre-trained embedding
                             methods (word2vec, glove, fastext)
            word_embed_dim: dimensionality of word embedding
            char_embed_type: same as word_embed_type, only apply when use_char is True
            char_embed_dim: dimensionality of char embedding
            external_word_dict: external word dictionary, only apply when use_word is True
            bert_vocab_file: vocabulary file of pre-trained bert model, only apply when use_bert is
                             True
            label_dict_file: a file with two columns separated by tab, the first column is raw
                             label name, and the second column is the corresponding name which is
                             meaningful
            max_len: int, max sequence length
            max_word_len: int, max word length
            padding_mode: str, 'pre' or 'post', pad either before or after each sequence
            truncating_mode: str, 'pre' or 'post', remove values from sequences larger than
                             `max_len`, either at the beginning or at the end of the sequences
        """
        super(SPMPreprocessor, self).__init__(max_len, padding_mode, truncating_mode)

        self.train_data = train_data
        self.train_labels = train_labels
        self.min_count = min_count
        self.use_word = use_word
        self.use_char = use_char
        self.use_bert = use_bert
        self.use_bert_model = use_bert_model
        self.external_word_dict = external_word_dict
        self.word_embed_type = word_embed_type
        self.char_embed_type = char_embed_type
        self.max_word_len = max_word_len

        self.label_dict = self.load_label_dict(label_dict_file)

        assert not (self.use_bert_model and (self.use_word or self.use_char)), \
            "bert model can not add word or char embedding as additional input"
        assert not (self.use_bert_model and not use_bert), "bert model must use bert embedding"
        assert self.use_word or self.use_char or self.use_bert, "must use word or char or bert" \
                                                                "embedding as main input"
        assert not (self.use_word and self.use_bert), "bert embedding can not be used with word" \
                                                      "embedding"
        special_token = 'bert' if self.use_bert else 'standard'

        train_data_a, train_data_b = self.train_data
        train_data = list(chain(*zip(train_data_a, train_data_b)))

        # build word vocabulary and word embedding
        if self.use_word:
            self.load_word_dict()

            word_corpus = self.build_corpus(train_data, cut_func=lambda x: jieba.lcut(x))
            self.word_vocab_count, self.word_vocab, self.id2word = \
                self.build_vocab(word_corpus, self.min_count, special_token)
            self.word_vocab_size = len(self.word_vocab)
            self.word_embeddings = self.build_embedding(word_embed_type, self.word_vocab,
                                                        word_corpus, word_embed_dim,
                                                        special_token)
            if self.word_embeddings is not None:
                self.word_embed_dim = self.word_embeddings.shape[1]
            else:
                self.word_embed_dim = word_embed_dim

            if self.max_len is None:
                self.max_len = get_len_from_corpus(word_corpus)
            if self.use_char and self.max_word_len is None:
                self.max_word_len = get_len_from_corpus(list(chain(*word_corpus)))
        else:
            self.word_vocab_count, self.word_vocab, self.id2word = None, None, None
            self.word_vocab_size = -1
            self.word_embeddings = None
            self.word_embed_dim = -1

        train_data = [list(text) for text in train_data]

        # build char vocabulary and char embedding
        if self.use_char:
            self.char_vocab_count, self.char_vocab, self.id2char = \
                self.build_vocab(train_data, self.min_count, special_token)
            self.char_vocab_size = len(self.char_vocab)
            self.char_embeddings = self.build_embedding(char_embed_type, self.char_vocab,
                                                        train_data, char_embed_dim,
                                                        special_token)
            if self.char_embeddings is not None:
                self.char_embed_dim = self.char_embeddings.shape[1]
            else:
                self.char_embed_dim = char_embed_dim
            if self.max_len is None:
                self.max_len = get_len_from_corpus(train_data)
        else:
            self.char_vocab_count, self.char_vocab, self.id2char = None, None, None
            self.char_vocab_size = -1
            self.char_embeddings = None
            self.char_embed_dim = -1

        # build bert vocabulary
        if self.use_bert:
            self.bert_vocab = {}
            with codecs.open(bert_vocab_file, 'r', 'utf8') as reader:
                for line in reader:
                    token = line.strip()
                    self.bert_vocab[token] = len(self.bert_vocab)
            self.bert_tokenizer = ChineseBertTokenizer(self.bert_vocab)

        # build label vocabulary
        self.label_vocab, self.id2label = self.build_label_vocab(self.train_labels)
        self.num_class = len(self.label_vocab)

        if self.use_bert_model and self.max_len is None:
            # max_len should be provided when use bert model!
            # We will reset max_len from train_data when max_len is not provided.
            self.max_len = get_len_from_corpus([list(a) + list(b) for a, b in zip(train_data_a,
                                                                                  train_data_b)])
        elif not self.use_word and self.use_bert and self.max_len is None:
            # max_len should be provided when use bert as input!
            # We will reset max_len from train_data when max_len is not provided.
            self.max_len = get_len_from_corpus(train_data)

        if self.use_bert:
            # max length is 512 for bert
            self.max_len = min(self.max_len, 512)

    def load_word_dict(self):
        if self.external_word_dict:
            for word in self.external_word_dict:
                jieba.add_word(word, freq=1000000)

    @staticmethod
    def load_label_dict(label_dict_file):
        result_dict = dict()
        if label_dict_file:
            with codecs.open(label_dict_file, encoding='utf-8') as f_label_dict:
                for line in f_label_dict:
                    line_items = line.strip().split('\t')
                    result_dict[line_items[0]] = line_items[1]
            return result_dict
        else:
            return None

    def build_label_vocab(self, labels):
        """Build label vocabulary

        Args:
            labels: list of str, the label strings
        """
        label_count = {}
        for label in labels:
            label_count[label] = label_count.get(label, 0) + 1

        # sorted by frequency, so that the label with the highest frequency will be given
        # id of 0, which is the default id for unknown labels
        sorted_label_count = sorted(label_count.items(), key=lambda x: x[1], reverse=True)
        sorted_label_count = dict(sorted_label_count)

        label_vocab = {}
        for label in sorted_label_count:
            label_vocab[label] = len(label_vocab)

        id2label = dict((idx, label) for label, idx in label_vocab.items())

        logging.info('Build label vocabulary finished, '
                     'vocabulary size: {}'.format(len(label_vocab)))
        return label_vocab, id2label

    def prepare_input(self, data, labels=None):
        """Prepare input (features and labels) for SPM model.
        Here we not only use character embeddings (or bert embeddings) as main input, but also
        support word embeddings and other hand-crafted features embeddings as additional input.

        Args:
            data: list of text pairs, like ``[['我是中国人', ...], ['我爱中国', ...]]``
            labels: list of str, the corresponding label strings

        Returns:
            features: id matrix
            y: label id matrix (only if labels is provided)

        """
        batch_word_ids_a, batch_char_ids_a, batch_bert_ids_a, batch_bert_seg_ids_a = \
            [], [], [], []
        batch_word_ids_b, batch_char_ids_b, batch_bert_ids_b, batch_bert_seg_ids_b = \
            [], [], [], []
        batch_label_ids = []

        for i, (text_a, text_b) in enumerate(zip(data[0], data[1])):
            if self.use_bert_model:
                indices, segments = self.bert_tokenizer.encode(first=text_a,
                                                               second=text_b,
                                                               max_len=self.max_len)
                batch_bert_ids_a.append(indices)
                batch_bert_seg_ids_a.append(segments)

            elif self.use_word:
                word_text_a = jieba.lcut(text_a)
                word_text_b = jieba.lcut(text_b)
                word_ids_a = self.get_word_ids(word_text_a)
                batch_word_ids_a.append(word_ids_a)
                word_ids_b = self.get_word_ids(word_text_b)
                batch_word_ids_b.append(word_ids_b)

                if self.use_char:
                    word_text_a = [list(word) for word in word_text_a]
                    word_text_b = [list(word) for word in word_text_b]
                    char_ids_a = [[self.char_vocab.get(char, self.char_vocab[self.unk_token])
                                   for char in token] for token in word_text_a]
                    char_ids_b = [[self.char_vocab.get(char, self.char_vocab[self.unk_token])
                                   for char in token] for token in word_text_b]
                    batch_char_ids_a.append(char_ids_a)
                    batch_char_ids_b.append(char_ids_b)

            else:
                text_a = list(text_a)
                text_b = list(text_b)

                if self.use_char:
                    char_text_a = [self.cls_token] + text_a + [self.seq_token] if self.use_bert \
                        else text_a
                    char_text_b = [self.cls_token] + text_b + [self.seq_token] if self.use_bert \
                        else text_b
                    char_ids_a = [self.char_vocab.get(token, self.char_vocab[self.unk_token])
                                  for token in char_text_a]
                    batch_char_ids_a.append(char_ids_a)
                    char_ids_b = [self.char_vocab.get(token, self.char_vocab[self.unk_token])
                                  for token in char_text_b]
                    batch_char_ids_b.append(char_ids_b)

                if self.use_bert:
                    indices_a, segments_a = self.bert_tokenizer.encode(first=''.join(text_a),
                                                                       max_len=self.max_len)
                    batch_bert_ids_a.append(indices_a)
                    batch_bert_seg_ids_a.append(segments_a)

                    indices_b, segments_b = self.bert_tokenizer.encode(first=''.join(text_b),
                                                                       max_len=self.max_len)
                    batch_bert_ids_b.append(indices_b)
                    batch_bert_seg_ids_b.append(segments_b)

            if labels is not None:
                label_ids = self.label_vocab.get(labels[i], self.get_unk_label_id())
                label_ids = to_categorical(label_ids, self.num_class).astype(int)
                batch_label_ids.append(label_ids)

        features_a, features_b = [], []
        if self.use_bert_model:
            features_a.append(self.pad_sequence(batch_bert_ids_a))
            features_a.append(self.pad_sequence(batch_bert_seg_ids_a))

        elif self.use_word:
            features_a.append(self.pad_sequence(batch_word_ids_a))
            features_b.append(self.pad_sequence(batch_word_ids_b))
            if self.use_char:
                features_a.append(pad_sequences_2d(batch_char_ids_a,
                                                   max_len_1=self.max_len,
                                                   max_len_2=self.max_word_len,
                                                   padding=self.padding_mode,
                                                   truncating=self.truncating_mode))
                features_b.append(pad_sequences_2d(batch_char_ids_b,
                                                   max_len_1=self.max_len,
                                                   max_len_2=self.max_word_len,
                                                   padding=self.padding_mode,
                                                   truncating=self.truncating_mode))

        else:
            if self.use_char:
                features_a.append(self.pad_sequence(batch_char_ids_a))
                features_b.append(self.pad_sequence(batch_char_ids_b))
            if self.use_bert:
                features_a.append(self.pad_sequence(batch_bert_ids_a))
                features_b.append(self.pad_sequence(batch_bert_ids_b))
                features_a.append(self.pad_sequence(batch_bert_seg_ids_a))
                features_b.append(self.pad_sequence(batch_bert_seg_ids_b))

        if len(features_a) == 1:
            features = [features_a[0], features_b[0]]
        else:
            features = features_a + features_b

        if not batch_label_ids:
            return features, None
        else:
            y = np.asarray(batch_label_ids)
            return features, y

    def get_word_ids(self, word_cut):
        """Given a word-level tokenized text, return the corresponding word ids.

        Args:
            word_cut: list of str, like ['我', '是'. '中国人']
            unk_idx: the index of words that do not appear in vocabulary, we usually set it to 1

        Returns: list of int, id sequence

        """
        word_ids = []
        for word in word_cut:
            word_ids.append(self.word_vocab.get(word, self.word_vocab[self.unk_token]))
        return word_ids

    def label_decode(self, pred_probs, label_dict=None):
        pred_ids = np.argmax(pred_probs, axis=-1)
        pred_labels = [self.id2label[pred_id] for pred_id in pred_ids]
        if label_dict:
            pred_labels = [label_dict[raw_label] for raw_label in pred_labels]
        return pred_labels

    def get_unk_label_id(self):
        """return a default id for label that does not exist in the label vocab

        Args:
            label: str

        Returns: int

        """
        if 'O' in self.label_vocab:
            return self.label_vocab['O']
        elif 'o' in self.label_vocab:
            return self.label_vocab['o']
        else:
            return 0

    def save(self, preprocessor_file):
        pickle.dump(self, open(preprocessor_file, 'wb'))

    @classmethod
    def load(cls, preprocessor_file):
        p = pickle.load(open(preprocessor_file, 'rb'))
        p.load_word_dict()  # reload external word dict into jieba
        return p



# -*- coding: utf-8 -*-

from keras.engine.topology import Layer


class NonMaskingLayer(Layer):
    """
    Fix convolutional 1D can't receive masked input.
    See: https://github.com/keras-team/keras/issues/4978
    Thanks for https://github.com/jacoxu
    """

    def __init__(self, **kwargs):
        self.supports_masking = True
        super(NonMaskingLayer, self).__init__(**kwargs)

    def build(self, input_shape):
        pass

    def compute_mask(self, inputs, input_mask=None):
        # do not pass the mask to the next layers
        return None

    def call(self, x, mask=None):
        return x



# -*- coding: utf-8 -*-

from absl import logging
from keras.models import model_from_json


def save_keras_model(model, json_file, weights_file):
    """Save keras model's architecture and weights to disk

    Args:
        model: keras model
        json_file: path to save model's architecture info
        weights_file: path to save model's weights

    Returns:

    """
    model_json = model.to_json()
    with open(json_file, 'w') as writer:
        writer.write(model_json)
    model.save_weights(weights_file)
    logging.info('Saved model to disk')


def load_keras_model(json_file, weights_file, custom_objects=None):
    """Load keras model from disk

    Args:
        json_file: file path to model's architecture info
        weights_file: file path to model's weights
        custom_objects: Optional dictionary mapping names (strings) to custom classes or
                        functions to be considered during deserialization. Must provided when
                        using custom layer
    """
    with open(json_file, 'r') as reader:
        model = model_from_json(reader.read(), custom_objects=custom_objects)
    model.load_weights(weights_file)
    return model



# -*- coding: utf-8 -*-

import numpy as np
from absl import logging


class TextClassificationPredictor(object):
    """TextClassification predictor for evaluating text classification model, output predictive
    probabilities and labels for input sentence"""
    def __init__(self, model, preprocessor):
        """

        Args:
            model: instance of keras model
            preprocessor: `TextClassificationPreprocessor` instance
        """
        self.model = model
        self.preprocessor = preprocessor

    def predict_prob(self, text):
        """Return probabilities for one sentence

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: np.array, shaped [num_classes,]

        """
        if isinstance(text, list):
            logging.warning('Text is passed in a list. Make sure it is tokenized in char level!')
            features, _ = self.preprocessor.prepare_input([text])
        else:
            assert isinstance(text, str)
            features, _ = self.preprocessor.prepare_input([list(text)])
        pred_probs = self.model.predict(features)
        return pred_probs[0]

    def predict_prob_batch(self, texts):
        """Return probabilities for a batch sentences

        Args:
            texts: a list of texts, each text can be untokenized (str) or
                   tokenized in char level (list)

        Returns: np.array, shaped [num_texts, num_classes]
        """
        assert isinstance(texts, list)
        if isinstance(texts[0], list):
            logging.warning('Text is passed in a list. Make sure it is tokenized in char level!')
            features, _ = self.preprocessor.prepare_input(texts)
        else:
            assert isinstance(texts[0], str)
            char_cut_texts = [list(text) for text in texts]
            features, _ = self.preprocessor.prepare_input(char_cut_texts)
        pred_probs = self.model.predict(features)
        return pred_probs

    def classify(self, text):
        """Return the classification result for one sentence

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: str

        """
        pred_prob = self.predict_prob(text)
        tags = self.preprocessor.label_decode(np.expand_dims(pred_prob, 0),
                                              self.preprocessor.label_dict)
        return tags[0]

    def classify_batch(self, texts):
        """Return classification result for a batch sentences

        Args:
            texts: a list of text, each text can be untokenized (str) or
                   tokenized in char level (list)

        Returns: list of str

        """
        pred_probs = self.predict_prob_batch(texts)
        tags = self.preprocessor.label_decode(pred_probs, self.preprocessor.label_dict)
        return tags

    def classification_with_prob(self, text):
        """Return the classification result for one sentence with probability

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: tuple

        """
        pred_result = self.predict_prob(text)
        tags = self.preprocessor.label_decode(np.expand_dims(pred_result, 0),
                                              self.preprocessor.label_dict)
        label_name = tags[0]
        label_prob = np.max(pred_result)
        return label_name, label_prob

    def classification_with_prob_batch(self, text):
        """Return the classification results for a batch sentences with probabilities

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: list of tuple

        """
        pred_results = self.predict_prob_batch(text)
        tags = self.preprocessor.label_decode(pred_results, self.preprocessor.label_dict)
        results = list(zip(tags, pred_results))
        return results



# -*- coding: utf-8 -*-

"""Base model with Keras
"""


class BaseModel(object):
    def build_input(self):
        """We build input and embedding layer for keras model here"""
        raise NotImplementedError

    def build_model(self):
        """We build keras model here"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
import codecs

from setuptools import setup
from setuptools import find_packages

long_description = '''
fancy-nlp is a fast and easy-to-use natural language processing (NLP) toolkit,
satisfying your imagination about NLP.

fancy-nlp is compatible with Python 3.6
and is distributed under the GPLv3 license.
'''

with codecs.open('requirements.txt', 'r', 'utf8') as reader:
    install_requires = list(map(lambda x: x.strip(), reader.readlines()))

setup(name='fancy-nlp',
      version='0.0.3',
      author='boat-group',
      author_email='e.shijia@foxmail.com',
      description='NLP for humans',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/boat-group/fancy-nlp",
      install_requires=install_requires,
      classifiers=[
          'Programming Language :: Python :: 3.6',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())




# -*- coding: utf-8 -*-

import math

import numpy as np
from keras.utils import Sequence


class NERGenerator(Sequence):
    """Data Generator for NER
    """
    def __init__(self, preprocessor, data, labels=None, batch_size=32, shuffle=True):
        """
        Args:
            preprocessor: `NERPreprocessor` instance to help prepare input for ner model
            data: list of tokenized texts (, like ``[['我', '是', '中', '国', '人']]``
            labels: list of list of str, the corresponding label strings
            batch_size: how many samples to train on in one iteration
            shuffle: whether to shuffle data after each epoch of training
        """
        self.preprocessor = preprocessor
        self.data = data
        self.labels = labels
        self.data_size = len(self.data)
        self.batch_size = batch_size
        self.indices = np.arange(self.data_size)
        self.steps = int(math.ceil(self.data_size / self.batch_size))
        self.shuffle = shuffle

    def __len__(self):
        return self.steps

    def on_epoch_end(self):
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __getitem__(self, index):
        batch_index = self.indices[index * self.batch_size: (index + 1) * self.batch_size]
        if self.labels is not None:
            batch_data, batch_labels = zip(*[(self.data[i], self.labels[i]) for i in batch_index])
        else:
            batch_data = [self.data[i] for i in batch_index]
            batch_labels = None
        return self.preprocessor.prepare_input(batch_data, batch_labels)


class TextClassificationGenerator(Sequence):
    """Data Generator for text classification
    """
    def __init__(self, preprocessor, data, labels=None, batch_size=32, shuffle=True):
        """
        Args:
            preprocessor: `TextClassificationPreprocessor` instance to help prepare input for ner model
            data: list of tokenized texts (, like ``[['我', '是', '中', '国', '人']]``
            labels: list of str, the corresponding label strings
            batch_size: how many samples to train on in one iteration
            shuffle: whether to shuffle data after each epoch of training
        """
        self.preprocessor = preprocessor
        self.data = data
        self.labels = labels
        self.data_size = len(self.data)
        self.batch_size = batch_size
        self.indices = np.arange(self.data_size)
        self.steps = int(math.ceil(self.data_size / self.batch_size))
        self.shuffle = shuffle

    def __len__(self):
        return self.steps

    def on_epoch_end(self):
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __getitem__(self, index):
        batch_index = self.indices[index * self.batch_size: (index + 1) * self.batch_size]
        if self.labels is not None:
            batch_data, batch_labels = zip(*[(self.data[i], self.labels[i]) for i in batch_index])
        else:
            batch_data = [self.data[i] for i in batch_index]
            batch_labels = None
        return self.preprocessor.prepare_input(batch_data, batch_labels)


class SPMGenerator(Sequence):
    """Data Generator for SPM
    """
    def __init__(self, preprocessor, data, labels=None, batch_size=32, shuffle=True):
        """
        Args:
            preprocessor: `SPMPreprocessor` instance to help prepare input for spm model
            data: list of text pairs (, like ``[['我是中国人', ...], ['我爱中国', ...]]``
            labels: list of str, the corresponding label strings
            batch_size: how many samples to train on in one iteration
            shuffle: whether to shuffle data after each epoch of training
        """
        self.preprocessor = preprocessor
        self.data = data
        self.labels = labels
        self.data_size = len(self.data[0])
        self.batch_size = batch_size
        self.indices = np.arange(self.data_size)
        self.steps = int(math.ceil(self.data_size / self.batch_size))
        self.shuffle = shuffle

    def __len__(self):
        return self.steps

    def on_epoch_end(self):
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __getitem__(self, index):
        batch_index = self.indices[index * self.batch_size: (index + 1) * self.batch_size]
        if self.labels is not None:
            batch_data_a, batch_data_b, batch_labels = \
                zip(*[(self.data[0][i], self.data[1][i], self.labels[i]) for i in batch_index])
        else:
            batch_data_a, batch_data_b = zip(*[(self.data[0][i], self.data[1][i])
                                               for i in batch_index])
            batch_labels = None
        return self.preprocessor.prepare_input((batch_data_a, batch_data_b), batch_labels)



# -*- coding: utf-8 -*-

from collections import defaultdict

import numpy as np
from absl import logging
from seqeval.metrics.sequence_labeling import get_entities


class NERPredictor(object):
    """NER predictor for evaluating ner model, output predictive probabilities and predictive tag
    sequences for input sentence"""
    def __init__(self, model, preprocessor):
        """

        Args:
            model: instance of keras model
            preprocessor: `NERPreprocessor` instance to prepare feature input for ner model
        """
        self.model = model
        self.preprocessor = preprocessor

    def predict_prob(self, text):
        """Return probabilities for one sentence

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: np.array, shaped [num_chars, num_classes]

        """
        if isinstance(text, list):
            logging.warning('Text is passed in a list. Make sure it is tokenized in char level!')
            features, _ = self.preprocessor.prepare_input([text])
        else:
            assert isinstance(text, str)
            features, _ = self.preprocessor.prepare_input([list(text)])
        pred_probs = self.model.predict(features)

        if self.preprocessor.use_bert:
            return pred_probs[0, 1:-1, :]
        else:
            return pred_probs[0]

    def predict_prob_batch(self, texts):
        """Return probabilities for a batch sentences

        Args:
            texts: a list of texts, each text can be untokenized (str) or
                   tokenized in char level (list)

        Returns: np.array, shaped [num_texts, num_chars, num_classes
        """
        assert isinstance(texts, list)
        if isinstance(texts[0], list):
            logging.warning('Text is passed in a list. Make sure it is tokenized in char level!')
            features, _ = self.preprocessor.prepare_input(texts)
        else:
            assert isinstance(texts[0], str)
            char_cut_texts = [list(text) for text in texts]
            features, _ = self.preprocessor.prepare_input(char_cut_texts)
        pred_probs = self.model.predict(features)

        if self.preprocessor.use_bert:
            return pred_probs[:, 1:-1, :]
        else:
            return pred_probs

    def tag(self, text):
        """Return tag sequence for one sentence

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns: list of str

        """
        pred_prob = self.predict_prob(text)
        length = min(len(text), pred_prob.shape[0])
        tags = self.preprocessor.label_decode(np.expand_dims(pred_prob, 0), [length])
        return tags[0]

    def tag_batch(self, texts):
        """Return tag sequences for a batch sentences

        Args:
            texts: a list of text, each text can be untokenized (str) or
                   tokenized in char level (list)

        Returns: list of list of str

        """
        pred_probs = self.predict_prob_batch(texts)
        lengths = [min(len(text), pred_prob.shape[0]) for text, pred_prob in zip(texts, pred_probs)]
        tags = self.preprocessor.label_decode(pred_probs, lengths)
        return tags

    @staticmethod
    def entities(text, tag, pred_prob):
        """Return entities according to tag sequence
        """
        results = []
        chunks = get_entities(tag)

        for chunk_type, chunk_start, chunk_end in chunks:
            chunk_end += 1
            entity = {
                'name': ''.join(text[chunk_start: chunk_end]),
                'type': chunk_type,
                'score': float(np.average(pred_prob[chunk_start: chunk_end])),
                'beginOffset': chunk_start,
                'endOffset': chunk_end
            }
            results.append(entity)
        return results

    @staticmethod
    def restrict_entities(text, tag, pred_prob, threshold=0.85):
        """Return restricted entities according to tag sequence: only keep at most one entity for
        each entity type
        """
        group_entities = defaultdict(list)

        chunks = get_entities(tag)
        for chunk_type, chunk_start, chunk_end in chunks:
            chunk_end += 1
            score = float(np.average(pred_prob[chunk_start: chunk_end]))
            if score >= threshold:
                entity = ''.join(text[chunk_start: chunk_end])
                group_entities[chunk_type].append((entity, score, chunk_start, chunk_end))

        results = []
        for entity_type, group in group_entities.items():
            entity = sorted(group, key=lambda x: x[0])[-1]
            results.append({
                'name': entity[0],
                'type': entity_type,
                'score': entity[1],
                'beginOffset': entity[2],
                'endOffset': entity[3]
            })
        return results

    def pretty_tag(self, text):
        """Return tag sequence for one sentence in a pretty format

        Args:
            text: can be untokenized (str) or tokenized in char level (list)

        Returns:

        """
        pred_prob = self.predict_prob(text)
        length = min(len(text), pred_prob.shape[0])
        tag = self.preprocessor.label_decode(np.expand_dims(pred_prob, 0), [length])

        pred_prob = np.max(pred_prob, axis=-1)
        char_cut = text if isinstance(text, list) else list(text)
        results = {
            'text': ''.join(char_cut),
            'entities': self.entities(char_cut, tag, pred_prob)
        }
        return results

    def pretty_tag_batch(self, texts):
        pred_probs = self.predict_prob_batch(texts)
        lengths = [min(len(text), pred_prob.shape[0]) for text, pred_prob in zip(texts, pred_probs)]
        tags = self.preprocessor.label_decode(pred_probs, lengths)

        pred_probs = np.max(pred_probs, axis=-1)
        results = []
        for text, tag, pred_prob in zip(texts, tags, pred_probs):
            char_cut = text if isinstance(text, list) else list(text)
            results.append({
                'text': ''.join(char_cut),
                'entities': self.entities(char_cut, tag, pred_prob)
            })
        return results

    def restrict_tag(self, text, threshold=0.85):
        """Return a restricted tag sequence for one sentence: only keep at most one entity
        for each entity type.

        Args:
            text: can be untokenized (str) or tokenized in char level (list)
            threshold:
        Returns:

        """
        pred_prob = self.predict_prob(text)
        length = min(len(text), pred_prob.shape[0])
        tag = self.preprocessor.label_decode(np.expand_dims(pred_prob, 0), [length])

        pred_prob = np.max(pred_prob, axis=-1)
        char_cut = text if isinstance(text, list) else list(text)
        results = {
            'text': ''.join(char_cut),
            'entities': self.restrict_entities(char_cut, tag, pred_prob, threshold)
        }
        return results

    def restrict_tag_batch(self, texts, threshold=0.85):
        pred_probs = self.predict_prob_batch(texts)
        lengths = [min(len(text), pred_prob.shape[0]) for text, pred_prob in zip(texts, pred_probs)]
        tags = self.preprocessor.label_decode(pred_probs, lengths)

        pred_probs = np.max(pred_probs, axis=-1)
        results = []
        for text, tag, pred_prob in zip(texts, tags, pred_probs):
            char_cut = text if isinstance(text, list) else list(text)
            results.append({
                'text': ''.join(char_cut),
                'entities': self.restrict_entities(char_cut, tag, pred_prob, threshold)
            })
        return results



# -*- coding: utf-8 -*-

from keras import initializers, regularizers, constraints
from keras.engine.topology import Layer
import keras.backend as K
import tensorflow as tf


class MultiHeadAttention(Layer):
    """
    Multi-head Attention introduced in Transformer, support masking
    """
    def __init__(self, num_units=100, num_heads=3, residual=True, normalize=True,
                 initializer='orthogonal', regularizer=None, constraint=None, **kwargs):
        self.num_units = num_units
        self.num_heads = num_heads
        self.model_units = self.num_units * self.num_heads
        self.residual = residual
        self.normalize = normalize
        self.initializer = initializers.get(initializer)
        self.regularizer = regularizers.get(regularizer)
        self.constraint = constraints.get(constraint)
        self.supports_masking = True
        super(MultiHeadAttention, self).__init__(**kwargs)

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise ValueError('Input into MultiHeadAttention should be a 3D input tensor')

        self.w_q = self.add_weight(name='w_q', shape=(input_shape[-1], self.model_units),
                                   initializer=self.initializer, regularizer=self.regularizer,
                                   constraint=self.constraint)
        self.w_k = self.add_weight(name='w_k', shape=(input_shape[-1], self.model_units),
                                   initializer=self.initializer, regularizer=self.regularizer,
                                   constraint=self.constraint)
        self.w_v = self.add_weight(name='w_v', shape=(input_shape[-1], self.model_units),
                                   initializer=self.initializer, regularizer=self.regularizer,
                                   constraint=self.constraint)
        self.w_final = self.add_weight(name='w_v', shape=(self.model_units, self.model_units),
                                       initializer=self.initializer, regularizer=self.regularizer,
                                       constraint=self.constraint)
        if self.normalize:
            self.gamma = self.add_weight(name='gamma', shape=(self.model_units,), initializer='one',
                                         regularizer=self.regularizer, constraint=self.constraint)
            self.beta = self.add_weight(name='beta', shape=(self.model_units,), initializer='zero',
                                        regularizer=self.regularizer, constraint=self.constraint)
        super(MultiHeadAttention, self).build(input_shape)

    def call(self, inputs, mask=None):
        """
        convert to query, key, value vectors, shaped [batch_size*num_head, time_step, embed_dim]
        """
        multihead_query = K.concatenate(tf.split(K.dot(inputs, self.w_q),
                                                 self.num_heads, axis=2), axis=0)
        multihead_key = K.concatenate(tf.split(K.dot(inputs, self.w_k),
                                               self.num_heads, axis=2), axis=0)
        multihead_value = K.concatenate(tf.split(K.dot(inputs, self.w_v),
                                                 self.num_heads, axis=2), axis=0)

        """scaled dot product"""
        scaled = K.int_shape(inputs)[-1] ** -0.5
        attend = K.batch_dot(multihead_query, multihead_key, axes=2) * scaled
        # apply mask before normalization (softmax)
        if mask is not None:
            multihead_mask = K.tile(mask, [self.num_heads, 1])
            attend *= K.expand_dims(K.cast(multihead_mask, K.floatx()), 2)
            attend *= K.expand_dims(K.cast(multihead_mask, K.floatx()), 1)
        # normalization
        attend = attend / K.cast(K.sum(attend, axis=-1, keepdims=True) + K.epsilon(), K.floatx())
        # apply attention
        attend = K.batch_dot(attend, multihead_value, axes=(2, 1))
        attend = tf.concat(tf.split(attend, self.num_heads, axis=0), axis=2)
        attend = K.dot(attend, self.w_final)

        if self.residual:
            attend = attend + inputs
        if self.normalize:
            mean = K.mean(attend, axis=-1, keepdims=True)
            std = K.mean(attend, axis=-1, keepdims=True)
            attend = self.gamma * (attend - mean) / (std + K.epsilon()) + self.beta

        return attend

    def compute_output_shape(self, input_shape):
        return input_shape[0], input_shape[1], self.num_units*self.num_heads

    def get_config(self):
        config = {'num_units': self.num_units,
                  'num_heads': self.num_heads,
                  'residual': self.residual,
                  'normalize': self.normalize,
                  'initializer': initializers.serialize(self.initializer),
                  'regularizer': regularizers.serialize(self.regularizer),
                  'constraint': constraints.serialize(self.constraint)}
        base_config = super(MultiHeadAttention, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



# -*- coding: utf-8 -*-

import pickle
import codecs

import numpy as np
import jieba
from absl import logging
from keras.utils.np_utils import to_categorical

from fancy_nlp.preprocessors.preprocessor import Preprocessor
from fancy_nlp.utils import pad_sequences_2d, get_len_from_corpus, ChineseBertTokenizer


class NERPreprocessor(Preprocessor):
    """NER preprocessor.
    """
    def __init__(self, train_data, train_labels, min_count=2, use_char=True, use_bert=False,
                 use_word=False, external_word_dict=None, bert_vocab_file=None,
                 char_embed_type=None, char_embed_dim=300, word_embed_type=None, word_embed_dim=300,
                 max_len=None, padding_mode='post', truncating_mode='post'):
        """

        Args:
            train_data: a list of tokenized (in char level) texts
            train_labels: list of list, train_data's labels
            min_count: int, token of which frequency is lower than min_count will be ignored
            use_char：whether to use char embedding as input
            use_bert: whether to use bert embedding as input
            use_word: whether to use word embedding as additional input
            external_word_dict: external word dictionary, only apply when use_word is True
            bert_vocab_file: vocabulary file of pre-trained bert model, only apply when use_bert is
                             True
            char_embed_type: str, can be a pre-trained embedding filename or pre-trained embedding
                             methods (word2vec, glove, fastext)
            char_embed_dim: dimensionality of char embedding
            word_embed_type: same as char_embed_type, only apply when use_word is True
            word_embed_dim: dimensionality of word embedding
            max_len: int, max sequence len
            padding_mode:
            truncating_mode:
        """
        super(NERPreprocessor, self).__init__(max_len, padding_mode, truncating_mode)

        self.train_data = train_data
        self.train_labels = train_labels
        self.min_count = min_count
        self.use_char = use_char
        self.use_bert = use_bert
        self.use_word = use_word
        self.external_word_dict = external_word_dict
        self.char_embed_type = char_embed_type
        self.word_embed_type = word_embed_type

        assert self.use_char or self.use_bert, "must use char or bert embedding as main input"
        special_token = 'bert' if self.use_bert else 'standard'

        # build char vocabulary and char embedding
        if self.use_char:
            self.char_vocab_count, self.char_vocab, self.id2char = \
                self.build_vocab(self.train_data, self.min_count, special_token)
            self.char_vocab_size = len(self.char_vocab)
            self.char_embeddings = self.build_embedding(char_embed_type, self.char_vocab,
                                                        self.train_data, char_embed_dim,
                                                        special_token)
            if self.char_embeddings is not None:
                self.char_embed_dim = self.char_embeddings.shape[1]
            else:
                self.char_embed_dim = char_embed_dim
        else:
            self.char_vocab_count, self.char_vocab, self.id2char = None, None, None
            self.char_vocab_size = -1
            self.char_embeddings = None
            self.char_embed_dim = -1

        # build bert vocabulary
        if self.use_bert:
            self.bert_vocab = {}
            with codecs.open(bert_vocab_file, 'r', 'utf8') as reader:
                for line in reader:
                    token = line.strip()
                    self.bert_vocab[token] = len(self.bert_vocab)
            self.bert_tokenizer = ChineseBertTokenizer(self.bert_vocab)

        # build word vocabulary and word embedding
        if self.use_word:
            self.load_word_dict()

            untokenized_texts = [''.join(text) for text in self.train_data]
            word_corpus = self.build_corpus(untokenized_texts, cut_func=lambda x: jieba.lcut(x))

            self.word_vocab_count, self.word_vocab, self.id2word = \
                self.build_vocab(word_corpus, self.min_count, special_token)
            self.word_vocab_size = len(self.word_vocab)
            self.word_embeddings = self.build_embedding(word_embed_type, self.word_vocab,
                                                        word_corpus, word_embed_dim,
                                                        special_token)
            if self.word_embeddings is not None:
                self.word_embed_dim = self.word_embeddings.shape[1]
            else:
                self.word_embed_dim = word_embed_dim
        else:
            self.word_vocab_count, self.word_vocab, self.id2word = None, None, None
            self.word_vocab_size = -1
            self.word_embeddings = None
            self.word_embed_dim = -1

        # build label vocabulary
        self.label_vocab, self.id2label = self.build_label_vocab(self.train_labels)
        self.num_class = len(self.label_vocab)

        if self.use_bert and self.max_len is None:
            # max_len must be provided when use bert as input!
            # We will reset max_len from train_data when max_len is not provided.
            self.max_len = get_len_from_corpus(self.train_data)

    def load_word_dict(self):
        if self.external_word_dict:
            for word in self.external_word_dict:
                jieba.add_word(word, freq=1000000)

    def build_label_vocab(self, labels):
        """Build label vocabulary

        Args:
            labels: list of list of str, the label strings
        """
        label_count = {}
        for sequence in labels:
            for label in sequence:
                label_count[label] = label_count.get(label, 0) + 1

        # sorted by frequency, so that the label with the highest frequency will be given
        # id of 0, which is the default id for unknown labels
        sorted_label_count = sorted(label_count.items(), key=lambda x: x[1], reverse=True)
        sorted_label_count = dict(sorted_label_count)

        label_vocab = {}
        for label in sorted_label_count:
            label_vocab[label] = len(label_vocab)

        id2label = dict((idx, label) for label, idx in label_vocab.items())

        logging.info('Build label vocabulary finished, '
                     'vocabulary size: {}'.format(len(label_vocab)))
        return label_vocab, id2label

    def prepare_input(self, data, labels=None):
        """Prepare input (features and labels) for NER model.
        Here we not only use character embeddings (or bert embeddings) as main input, but also
        support word embeddings and other hand-crafted features embeddings as additional input.

        Args:
            data: list of tokenized (in char level) texts, like ``[['我', '是', '中', '国', '人']]``
            labels: list of list of str, the corresponding label strings

        Returns:
            features: id matrix
            y: label id matrix (only if labels is provided)

        """
        batch_char_ids, batch_bert_ids, batch_bert_seg_ids, batch_word_ids = [], [], [], []
        batch_label_ids = []
        for i, char_text in enumerate(data):
            if self.use_char:
                if self.use_bert:
                    text_for_char_input = [self.cls_token] + char_text + [self.seq_token]
                else:
                    text_for_char_input = char_text
                char_ids = [self.char_vocab.get(token, self.char_vocab[self.unk_token])
                            for token in text_for_char_input]
                batch_char_ids.append(char_ids)

            if self.use_bert:
                indices, segments = self.bert_tokenizer.encode(first=''.join(char_text),
                                                               max_len=self.max_len)
                batch_bert_ids.append(indices)
                batch_bert_seg_ids.append(segments)

            if self.use_word:
                word_text = jieba.lcut(''.join(char_text))
                word_ids = self.get_word_ids(word_text)
                batch_word_ids.append(word_ids)

            if labels is not None:
                if self.use_bert:
                    label_str = [self.cls_token] + labels[i] + [self.cls_token]
                else:
                    label_str = labels[i]
                label_ids = [self.label_vocab.get(l, self.get_unk_label_id()) for l in label_str]
                label_ids = to_categorical(label_ids, self.num_class).astype(int)
                batch_label_ids.append(label_ids)

        features = []
        if self.use_char:
            features.append(self.pad_sequence(batch_char_ids))
        if self.use_bert:
            features.append(self.pad_sequence(batch_bert_ids))
            features.append(self.pad_sequence(batch_bert_seg_ids))
        if self.use_word:
            features.append(self.pad_sequence(batch_word_ids))

        if len(features) == 1:
            features = features[0]

        if not batch_label_ids:
            return features, None
        else:
            y = pad_sequences_2d(batch_label_ids, max_len_1=self.max_len, max_len_2=self.num_class,
                                 padding=self.padding_mode, truncating=self.truncating_mode)
            return features, y

    def get_word_ids(self, word_cut):
        """Given a word-level tokenized text, return the corresponding word ids in char-level
           sequence. We add the same word id to each character in the word.

        Args:
            word_cut: list of str, like ['我', '是'. '中国人']
            unk_idx: the index of words that do not appear in vocabulary, we usually set it to 1

        Returns: list of int, id sequence

        """
        word_ids = []
        for word in word_cut:
            for _ in word:
                word_ids.append(self.word_vocab.get(word, self.word_vocab[self.unk_token]))
        if self.use_bert:
            word_ids = [self.word_vocab[self.cls_token]] + word_ids + \
                       [self.word_vocab[self.seq_token]]
        return word_ids

    def label_decode(self, pred_probs, lengths=None):
        pred_ids = np.argmax(pred_probs, axis=-1)
        pred_labels = [[self.id2label[label_id] for label_id in ids] for ids in pred_ids]
        if lengths is not None:
            pred_labels = [labels[:length] for labels, length in zip(pred_labels, lengths)]
        return pred_labels

    def get_unk_label_id(self):
        """return a default id for label that does not exist in the label vocab

        Args:
            label: str

        Returns: int

        """
        if 'O' in self.label_vocab:
            return self.label_vocab['O']
        elif 'o' in self.label_vocab:
            return self.label_vocab['o']
        else:
            return 0

    def save(self, preprocessor_file):
        pickle.dump(self, open(preprocessor_file, 'wb'))

    @classmethod
    def load(cls, preprocessor_file):
        p = pickle.load(open(preprocessor_file, 'rb'))
        p.load_word_dict()  # reload external word dict into jieba
        return p
